# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Manipulates the code to apply migrations."""

import logging

from libcst import (
    Annotation,
    Arg,
    Attribute,
    Call,
    CSTTransformer,
    Import,
    ImportAlias,
    ImportFrom,
    ImportStar,
    Name,
    SimpleString,
    ensure_type,
    metadata,
)
from libcst.helpers import get_full_name_for_node


class MigrationTransformer(CSTTransformer):
    """Manipulates the libcst CST to perform code migrations."""

    METADATA_DEPENDENCIES = (metadata.PositionProvider,)

    def __init__(self, filename, recording, mappings):
        """Construct a new transformer."""
        self.filename = filename
        self.recording = recording
        self.mappings = mappings
        self.call_stack = []

    def leave_Arg(self, original_node: Arg, updated_node: Arg) -> Arg:  # noqa: N802, D102
        # The keyword of Arg is neither an Assignment nor an Access and we explicitly don't visit it.
        if updated_node.keyword is not None:
            function_call = self.call_stack[-1]
            member_name = function_call.split(".")[-1]
            named_argument_name = updated_node.keyword.value
            logging.debug(f"Leaving argument: {function_call}({named_argument_name})")

            pos = self.get_metadata(metadata.PositionProvider, original_node)

            change = self.recording.get_change_for_file_region(
                self.filename,
                member_name,
                pos.start.line,
                pos.end.line,
                pos.start.column,
                pos.end.column,
            )

            if change is not None:
                if (
                    new_named_argument_name := self.mappings.get_mapping_for_named_argument(
                        change.type_name + "." + member_name,
                        named_argument_name,
                    )
                ) and new_named_argument_name is not None:
                    updated_node = updated_node.with_changes(keyword=Name(new_named_argument_name))

        return updated_node

    def visit_Call(self, node: Call) -> None:  # noqa: N802, D102
        full_name = get_full_name_for_node(node)
        self.call_stack.append(full_name)

    def leave_Call(self, original_node: Call, updated_node: Call) -> Call:  # noqa: N802, D102

        self.call_stack.pop()

        # Intercept constructor calls
        full_name = get_full_name_for_node(updated_node)

        if full_name is None:
            return updated_node

        full_name_parts = full_name.split(".")
        qualifier = ".".join(full_name_parts[:-1])
        name = full_name_parts[-1]

        if self.mappings.has_mapping_for_type(name):
            pos = self.get_metadata(metadata.PositionProvider, original_node)

            change = self.recording.get_change_for_file_region(
                self.filename,
                "__init__",
                pos.start.line,
                pos.end.line,
                pos.start.column,
                pos.end.column,
            )

            if change is not None:
                new_name = self.mappings.get_replacement_for_type(None, name)
                new_qualifier = self.mappings.get_replacement_for_namespace(qualifier)
                updated_node = updated_node.with_deep_changes(
                    updated_node,
                    func=self._assemble_names([x for x in new_qualifier.split(".") if x] + [new_name]),
                )
                return updated_node

        return updated_node

    def leave_Name(self, original_node, updated_node):  # noqa: N802, D102
        name = original_node.value

        if name is not None and self.mappings.is_member_name_of_interest(name):
            pos = self.get_metadata(metadata.PositionProvider, original_node)

            change = self.recording.get_change_for_file_region(
                self.filename,
                name,
                pos.start.line,
                pos.end.line,
                pos.start.column,
                pos.end.column,
            )

            if change is not None:
                new_name = self.mappings.get_replacement_for_member(change.type_name, change.member_name)
                logging.debug(f"Leaving name: {name}: {change is not None} - {pos} -> {new_name}")
                updated_node = updated_node.with_changes(value=new_name)
            else:
                logging.debug(f"Leaving name: {name}: {change is not None} - {pos} -> not found")

        return updated_node

    def leave_Attribute(self, original_node, updated_node):  # noqa: N802, D102

        attr_node = ensure_type(updated_node, Attribute)
        if isinstance(attr_node.value, Name) and isinstance(attr_node.attr, Name):
            lhs_name = attr_node.value.value
            rhs_name = attr_node.attr.value
            (new_enum_type, new_enum_value) = self.mappings.get_replacement_for_enum(lhs_name, rhs_name)
            logging.debug(f"Leaving attribute: {lhs_name}.{rhs_name} -> {new_enum_type}.{new_enum_value}")
            updated_node = updated_node.with_changes(value=Name(new_enum_type), attr=Name(new_enum_value))

        return updated_node

    def _assemble_names(self, names):
        if len(names) == 1:
            return Name(value=names[0])
        else:
            return Attribute(
                value=self._assemble_names(names[:-1]),
                attr=Name(value=names[-1]),
            )

    def leave_Import(self, original_node: Import, updated_node: Import) -> None:  # noqa: N802, D102

        module_name = updated_node.names

        new_names = []
        for n in module_name:
            old_imported_module = get_full_name_for_node(n.name)
            new_imported_module = self.mappings.get_replacement_for_namespace(old_imported_module)
            new_names.append(
                ImportAlias(
                    name=self._assemble_names(new_imported_module.split(".")),
                    asname=n.asname,
                )
            )

        updated_node = updated_node.with_changes(names=new_names)

        return updated_node

    def leave_Annotation(self, original_node: Annotation, updated_node: Annotation) -> None:  # noqa: N802, D102

        if isinstance(updated_node.annotation, SimpleString):
            was_quoted = True
            annotation_full_name = updated_node.annotation.value[1:-1]  # remove quotes
        else:
            was_quoted = False
            annotation_full_name = get_full_name_for_node(updated_node.annotation)

        logging.debug(f"Type hint found: {annotation_full_name}")

        full_name_parts = annotation_full_name.split(".")
        qualifier = ".".join(full_name_parts[:-1])
        name = full_name_parts[-1]
        new_type_name = self.mappings.get_replacement_for_type(qualifier, name)

        new_qualifier = self.mappings.get_replacement_for_namespace(qualifier)

        if was_quoted:
            updated_node = updated_node.with_changes(
                annotation=SimpleString(
                    '"' + ".".join([x for x in new_qualifier.split(".") if x] + [new_type_name]) + '"'
                )
            )
        else:
            updated_node = updated_node.with_changes(
                annotation=self._assemble_names([x for x in new_qualifier.split(".") if x] + [new_type_name])
            )

        return updated_node

    def leave_ImportFrom(self, original_node: ImportFrom, updated_node: ImportFrom) -> None:  # noqa: N802, D102

        module_name = updated_node.module
        old_imported_module = get_full_name_for_node(module_name)

        new_imported_module = self.mappings.get_replacement_for_namespace(old_imported_module)

        old_names = updated_node.names
        new_names = []
        if not isinstance(old_names, ImportStar):
            for old_name in old_names:
                old_imported_name = old_name.name.value
                new_names.append(
                    ImportAlias(
                        name=Name(self.mappings.get_replacement_for_type(old_imported_module, old_imported_name)),
                        asname=old_name.asname,
                    )
                )

        updated_node = updated_node.with_deep_changes(
            updated_node,
            module=self._assemble_names(new_imported_module.split(".")),
        )

        if len(new_names) > 0:
            updated_node = updated_node.with_deep_changes(
                updated_node,
                names=new_names,
            )

        return updated_node



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import typing

from enum     import IntEnum
from ctypes   import byref, cast, pointer, POINTER
from datetime import datetime, timedelta
from math     import floor, ceil

from .           import comutil     as agcom
from .           import coclassutil as agcoclass
from ..utilities import colors      as agcolor
from ..utilities import comobject   as agcomobj

from ..utilities.exceptions import STKColorError, STKPluginMethodNotImplementedError

###############################################################################
#   Marshalling DATE
###############################################################################
def datetime_to_DATE(d:datetime) -> agcom.DATE:
    delta = d-datetime(1899, 12, 30, 0, 0, 0)
    if delta.total_seconds() >= 0:
        return agcom.DATE(delta.total_seconds()/60/60/24)
    elif delta.total_seconds() > -1.0:
        return agcom.DATE(abs(delta.total_seconds())/60/60/24)
    else:
        days_before = ceil(abs(delta.total_seconds()/60/60/24))
        fraction_into_day = days_before - abs(delta.total_seconds()/60/60/24)
        value = -1*days_before - fraction_into_day
        return agcom.DATE(value)

def DATE_to_datetime(d:agcom.DATE) -> datetime:
    if d.value >= 0:
        delta = timedelta(days = d.value)
        return datetime(1899, 12, 30, 0, 0, 0) + delta
    else:
        days = floor(abs(d.value))
        seconds_into_day = (abs(d.value) - floor(abs(d.value)))*60*60*24
        delta_days = timedelta(days = days)
        delta_secs = timedelta(seconds = seconds_into_day)
        return (datetime(1899, 12, 30, 0, 0, 0) - delta_days) + delta_secs


###############################################################################
#   Marshalling Variant
###############################################################################
def pytype_to_vartype(item: typing.Any) -> int:
    python_type = type(item)
    if python_type==bool:
        return agcom.VT_BOOL
    elif python_type==str:
        return agcom.VT_BSTR
    elif python_type==float:
        return agcom.VT_R8
    elif python_type==int:
        if item > agcom.ULLONG_MAX:
            return agcom.VT_R8
        elif item > agcom.LLONG_MAX:
            return agcom.VT_UI8
        elif item > agcom.LONG_MAX:
            return agcom.VT_I8
        elif item > agcom.LONG_MIN:
            return agcom.VT_I4
        elif item > agcom.LLONG_MIN:
            return agcom.VT_I8
        else:
            return agcom.VT_R8
    elif python_type==list:
        if len(item)>0:
            vt = pytype_to_vartype(item[0])
            return agcom.VT_ARRAY|vt
        else:
            return agcom.VT_ARRAY|agcom.VT_NULL
    elif hasattr(item, "_intf") and type(item._intf)==agcom.IUnknown:
        return agcom.VT_UNKNOWN

def VARIANT_from_python_data(data:typing.Any) -> agcom.Variant:
    var = agcom.Variant()
    var.vt = agcom.VT_EMPTY
    if data is not None:
        if hasattr(data, "_to_argb"):
            data = data._to_argb()
        var.vt = pytype_to_vartype(data)
        union_val = agcom.varUnion()
        if var.vt == agcom.VT_BOOL:
            union_val.boolVal = agcom.VARIANT_TRUE if data else agcom.VARIANT_FALSE
        elif var.vt == agcom.VT_BSTR:
            union_val.bstrVal = agcom.BSTR(agcom.OLEAut32Lib.SysAllocString(data))
        elif var.vt == agcom.VT_I4:
            union_val.lVal = agcom.LONG(data)
        elif var.vt == agcom.VT_I8:
            union_val.llVal = agcom.LONGLONG(data)
        elif var.vt == agcom.VT_UI8:
            union_val.ullVal = agcom.ULONGLONG(data)
        elif var.vt == agcom.VT_R8:
            union_val.dblVal = agcom.DOUBLE(data)
        elif var.vt == agcom.VT_UNKNOWN:
            union_val.punkVal = data._intf.p
            agcom._CreateAgObjectLifetimeManager._add_ref_impl(data._intf)
        elif var.vt & agcom.VT_ARRAY:
            union_val.parray = SAFEARRAY_from_list(data, True)
        var.buffer = union_val.buffer
    return var

def ctype_val_from_VARIANT(var:agcom.Variant) -> typing.Any:
    union_val = agcom.varUnion()
    union_val.buffer = var.buffer
    if var.vt == agcom.VT_I8:
        return agcom.LONGLONG(union_val.llVal)
    elif var.vt == agcom.VT_I4:
        return agcom.LONG(union_val.lVal)
    elif var.vt == agcom.VT_UI1:
        return agcom.BYTE(union_val.bVal)
    elif var.vt == agcom.VT_I2:
        return agcom.SHORT(union_val.iVal)
    elif var.vt == agcom.VT_R4:
        return agcom.FLOAT(union_val.fltVal)
    elif var.vt == agcom.VT_R8:
        return agcom.DOUBLE(union_val.dblVal)
    elif var.vt == agcom.VT_BOOL:
        return agcom.VARIANT_BOOL(union_val.boolVal)
    elif var.vt == agcom.VT_BSTR:
        return agcom.BSTR(union_val.bstrVal)
    elif var.vt == agcom.VT_UNKNOWN:
        return agcom.PVOID(union_val.punkVal)
    elif var.vt & agcom.VT_ARRAY:
        if var.vt & agcom.VT_BYREF:
            return union_val.pparray
        else:
            return union_val.parray
    elif var.vt == agcom.VT_UI1|agcom.VT_BYREF:
        return union_val.pbVal
    elif var.vt == agcom.VT_I2|agcom.VT_BYREF:
        return union_val.piVal
    elif var.vt == agcom.VT_I4|agcom.VT_BYREF:
        return union_val.plVal
    elif var.vt == agcom.VT_I8|agcom.VT_BYREF:
        return union_val.pllVal
    elif var.vt == agcom.VT_R4|agcom.VT_BYREF:
        return union_val.pfltVal
    elif var.vt == agcom.VT_R8|agcom.VT_BYREF:
        return union_val.pdblVal
    elif var.vt == agcom.VT_BOOL|agcom.VT_BYREF:
        return union_val.pboolVal
    elif var.vt == agcom.VT_BSTR|agcom.VT_BYREF:
        return union_val.pbstrVal
    elif var.vt == agcom.VT_UNKNOWN|agcom.VT_BYREF:
        return union_val.ppunkVal
    elif var.vt == agcom.VT_I1:
        return agcom.CHAR(union_val.cVal)
    elif var.vt == agcom.VT_UI2:
        return agcom.USHORT(union_val.uiVal)
    elif var.vt == agcom.VT_UI4:
        return agcom.ULONG(union_val.ulVal)
    elif var.vt == agcom.VT_UI8:
        return agcom.ULONGLONG(union_val.ullVal)
    elif var.vt == agcom.VT_INT:
        return agcom.INT(union_val.intVal)
    elif var.vt == agcom.VT_UINT:
        return agcom.UINT(union_val.uintVal)
    elif var.vt == agcom.VT_I1|agcom.VT_BYREF:
        return union_val.pcVal
    elif var.vt == agcom.VT_UI2|agcom.VT_BYREF:
        return union_val.piVal
    elif var.vt == agcom.VT_UI4|agcom.VT_BYREF:
        return union_val.plVal
    elif var.vt == agcom.VT_UI8|agcom.VT_BYREF:
        return union_val.pllVal
    elif var.vt == agcom.VT_INT|agcom.VT_BYREF:
        return union_val.pintVal
    elif var.vt == agcom.VT_UINT|agcom.VT_BYREF:
        return union_val.puintVal
    elif var.vt == agcom.VT_VARIANT|agcom.VT_BYREF:
        return cast(union_val.pvarVal, POINTER(agcom.Variant))
    elif var.vt == agcom.VT_ERROR:
        return agcom.HRESULT(union_val.scode)
    elif var.vt == agcom.VT_ERROR|agcom.VT_BYREF:
        return union_val.pscode
    elif var.vt == agcom.VT_DISPATCH:
        return agcom.PVOID(union_val.pdispVal)
    elif var.vt == agcom.VT_DISPATCH|agcom.VT_BYREF:
        return union_val.ppdispVal

def vartype_is_integral_type(vt:int) -> bool:
    return vt in [agcom.VT_I1,
                  agcom.VT_I2,
                  agcom.VT_I4,
                  agcom.VT_I8,
                  agcom.VT_UI1,
                  agcom.VT_UI2,
                  agcom.VT_UI4,
                  agcom.VT_UI8,
                  agcom.VT_INT,
                  agcom.VT_UINT,
                  agcom.VT_ERROR]

def python_val_from_ctypes_val(ctypes_val:typing.Any, vt:int):
    if vt == agcom.VT_BOOL or vt == agcom.VT_BOOL|agcom.VT_BYREF:
        if vt & agcom.VT_BYREF:
            val = ctypes_val.contents.value
        else:
            val = ctypes_val.value
        return False if val == 0 else True
    elif vartype_is_integral_type(vt):
        return ctypes_val.value
    elif vartype_is_integral_type(vt ^ agcom.VT_BYREF):
        return ctypes_val.contents.value
    elif vt in [agcom.VT_R4, agcom.VT_R8]:
        return ctypes_val.value
    elif vt ^ agcom.VT_BYREF in [agcom.VT_R4, agcom.VT_R8]:
        return ctypes_val.contents.value
    elif vt == agcom.VT_BSTR or vt == agcom.VT_BSTR|agcom.VT_BYREF:
        if vt & agcom.VT_BYREF:
            val = ctypes_val.contents.value
        else:
            val = ctypes_val.value
        return val
    elif vt == agcom.VT_UNKNOWN or vt == agcom.VT_UNKNOWN|agcom.VT_BYREF \
        or vt == agcom.VT_DISPATCH or vt == agcom.VT_DISPATCH|agcom.VT_BYREF:
        if vt & agcom.VT_BYREF:
            pUnk = agcom.PVOID(ctypes_val.contents.value)
        else:
            pUnk = ctypes_val
        ret = agcom.IUnknown()
        ret.p = pUnk
        ret.create_ownership()
        return agcoclass.get_concrete_class(ret)
    elif vt & agcom.VT_ARRAY:
        if vt & agcom.VT_BYREF:
            sa = ctypes_val.contents
        else:
            sa = ctypes_val
        return list_from_SAFEARRAY(sa)
    elif vt == agcom.VT_VARIANT or vt == agcom.VT_VARIANT|agcom.VT_BYREF:
        if vt & agcom.VT_BYREF:
            return python_val_from_VARIANT(ctypes_val.contents)
        else:
            return python_val_from_VARIANT(ctypes_val)
    elif vt in [agcom.VT_EMPTY, agcom.VT_NULL]:
        return None
    raise RuntimeError("Unrecognized variant type: " + str(vt))

def python_val_from_VARIANT(var:agcom.Variant, clear_variant:bool=False) -> typing.Any:
    retval = python_val_from_ctypes_val(ctype_val_from_VARIANT(var), var.vt)
    if clear_variant:
        agcom.OLEAut32Lib.VariantClear(var)
    return retval

###############################################################################
#   Marshalling SAFEARRAY
###############################################################################
def SAFEARRAY_elem_from_python_elem(python_elem:typing.Any, as_VARIANT:bool = True) -> typing.Any:
    if as_VARIANT:
        return VARIANT_from_python_data(python_elem)
    else:
        t = type(python_elem)
        if t == str:
            # IAgCatAny expects SAFEARRAY(BSTR)
            return agcom.BSTR(agcom.SysAllocString(python_elem))
        else:
            raise RuntimeError("Unexpected Safearray element type.")

def _create_SAFEARRAY(vt:int, dim:int, num_elems_dim1:int, num_elems_dim2:int = 0) -> agcom.LPSAFEARRAY:
    rgsabound = (agcom.SafearrayBound*dim)()
    num_elems = [num_elems_dim1]
    if dim == 2:
        num_elems.append(num_elems_dim2)
    for i in range(dim):
        rgsabound[i].lLbound = agcom.LONG(0)
        rgsabound[i].cElements = agcom.ULONG(num_elems[i])
    sa = agcom.LPSAFEARRAY(agcom.OLEAut32Lib.SafeArrayCreate(vt, agcom.UINT(dim), cast(pointer(rgsabound), POINTER(agcom.SafearrayBound))))
    return sa

def SAFEARRAY_from_list(data:list, as_VARIANT:bool=True) -> agcom.LPSAFEARRAY:
    if len(data) == 0:
        return _create_SAFEARRAY(agcom.VT_VARIANT, 1, 0)
    elif type(data[0]) == list:
        #have a 2-D array; always as_VARIANT
        indices = (agcom.LONG*2)()
        indices[0] = 0
        indices[1] = 0
        pIndx = cast(pointer(indices), POINTER(agcom.LONG))
        sa = _create_SAFEARRAY(agcom.VT_VARIANT, 2, len(data), len(data[0]))
        for i in range(len(data)):
            for j in range(len(data[0])):
                indices[0] = agcom.LONG(i)
                indices[1] = agcom.LONG(j)
                elem = SAFEARRAY_elem_from_python_elem(data[i][j])
                agcoclass.evaluate_hresult(agcom.OLEAut32Lib.SafeArrayPutElement(sa, pIndx, byref(elem)))
        return sa
    else:
        # 1-D Vector
        vt = agcom.VT_VARIANT
        if not as_VARIANT and type(data[0]) == str:
            vt = agcom.VT_BSTR
        sa = _create_SAFEARRAY(vt, 1, len(data))
        for i in range(len(data)):
            index = agcom.LONG(i)
            elem = SAFEARRAY_elem_from_python_elem(data[i], as_VARIANT)
            if vt == agcom.VT_BSTR:
                agcom.OLEAut32Lib.SafeArrayPutElement(sa, byref(index), elem)
            else:
                agcom.OLEAut32Lib.SafeArrayPutElement(sa, byref(index), byref(elem))
        return sa

def _vartype_to_ctypes_type(vt:agcom.INT) -> typing.Any:
    if vt == agcom.VT_I8:
        return agcom.LONGLONG
    elif vt == agcom.VT_I4:
        return agcom.LONG
    elif vt == agcom.VT_UI1:
        return agcom.BYTE
    elif vt == agcom.VT_I2:
        return agcom.SHORT
    elif vt == agcom.VT_R4:
        return agcom.FLOAT
    elif vt == agcom.VT_R8:
        return agcom.DOUBLE
    elif vt == agcom.VT_BOOL:
        return agcom.VARIANT_BOOL
    elif vt == agcom.VT_BSTR:
        return agcom.BSTR
    elif vt == agcom.VT_UNKNOWN:
        return agcom.PVOID
    elif vt & agcom.VT_ARRAY:
        return agcom.PVOID
    elif vt == agcom.VT_UI1|agcom.VT_BYREF:
        return POINTER(agcom.BYTE)
    elif vt == agcom.VT_I2|agcom.VT_BYREF:
        return POINTER(agcom.SHORT)
    elif vt == agcom.VT_I4|agcom.VT_BYREF:
        return POINTER(agcom.LONG)
    elif vt == agcom.VT_I8|agcom.VT_BYREF:
        return POINTER(agcom.LONGLONG)
    elif vt == agcom.VT_R4|agcom.VT_BYREF:
        return POINTER(agcom.FLOAT)
    elif vt == agcom.VT_R8|agcom.VT_BYREF:
        return POINTER(agcom.DOUBLE)
    elif vt == agcom.VT_BOOL|agcom.VT_BYREF:
        return POINTER(agcom.VARIANT_BOOL)
    elif vt == agcom.VT_BSTR|agcom.VT_BYREF:
        return POINTER(agcom.BSTR)
    elif vt == agcom.VT_UNKNOWN|agcom.VT_BYREF:
        return POINTER(agcom.PVOID)
    elif vt == agcom.VT_I1:
        return agcom.CHAR
    elif vt == agcom.VT_UI2:
        return agcom.USHORT
    elif vt == agcom.VT_UI4:
        return agcom.ULONG
    elif vt == agcom.VT_UI8:
        return agcom.ULONGLONG
    elif vt == agcom.VT_INT:
        return agcom.INT
    elif vt == agcom.VT_UINT:
        return agcom.UINT
    elif vt == agcom.VT_I1|agcom.VT_BYREF:
        return POINTER(agcom.CHAR)
    elif vt == agcom.VT_UI2|agcom.VT_BYREF:
        return POINTER(agcom.USHORT)
    elif vt == agcom.VT_UI4|agcom.VT_BYREF:
        return POINTER(agcom.ULONG)
    elif vt == agcom.VT_UI8|agcom.VT_BYREF:
        return POINTER(agcom.ULONGLONG)
    elif vt == agcom.VT_INT|agcom.VT_BYREF:
        return POINTER(agcom.INT)
    elif vt == agcom.VT_UINT|agcom.VT_BYREF:
        return POINTER(agcom.UINT)
    elif vt == agcom.VT_VARIANT:
        return agcom.Variant
    elif vt == agcom.VT_VARIANT|agcom.VT_BYREF:
        return POINTER(agcom.Variant)
    elif vt == agcom.VT_ERROR:
        return agcom.HRESULT
    elif vt == agcom.VT_ERROR|agcom.VT_BYREF:
        return POINTER(agcom.HRESULT)
    elif vt == agcom.VT_DISPATCH:
        return agcom.PVOID
    elif vt == agcom.VT_DISPATCH|agcom.VT_BYREF:
        return POINTER(agcom.PVOID)
    else:
        raise RuntimeError("Unrecognized variant type: " + str(vt))

def _single_dimension_list_from_SAFEARRAY(sa:agcom.LPSAFEARRAY, index:int, from_2d_array=False) -> list:
    python_array = list()
    vt = agcom.VARTYPE()
    agcom.OLEAut32Lib.SafeArrayGetVartype(sa, byref(vt))
    lb1 = agcom.LONG()
    ub1 = agcom.LONG()
    agcom.OLEAut32Lib.SafeArrayGetLBound(sa, agcom.UINT(1), byref(lb1))
    agcom.OLEAut32Lib.SafeArrayGetUBound(sa, agcom.UINT(1), byref(ub1))
    nElem1 = int(ub1.value)+1-int(lb1.value)
    pVoid = agcom.LPVOID()
    hr = agcom.OLEAut32Lib.SafeArrayAccessData(sa, byref(pVoid))
    pElem = cast(pVoid, POINTER(_vartype_to_ctypes_type(vt.value)))
    if not from_2d_array:
        for i in range(int(lb1.value), int(lb1.value)+nElem1):
            python_array.append(python_val_from_ctypes_val(pElem[i], vt.value))
    else:
        lb2 = agcom.LONG()
        ub2 = agcom.LONG()
        agcom.OLEAut32Lib.SafeArrayGetLBound(sa, agcom.UINT(2), byref(lb2))
        agcom.OLEAut32Lib.SafeArrayGetUBound(sa, agcom.UINT(2), byref(ub2))
        nElem2 = int(ub2.value)+1-int(lb2.value)
        for i in range(int(lb2.value), int(lb2.value)+nElem2):
            python_array.append(python_val_from_ctypes_val(pElem[(i * nElem1) + index], vt.value))
    hr = agcom.OLEAut32Lib.SafeArrayUnaccessData(sa)
    return python_array


def _five_dimension_list_from_SAFEARRAY(sa:agcom.LPSAFEARRAY) -> list:
    python_array = []
    vt = agcom.VARTYPE()
    agcom.OLEAut32Lib.SafeArrayGetVartype(sa, byref(vt))
    if vt.value != agcom.VT_VARIANT:
        return []
    indices = (agcom.LONG*5)()
    indices[0] = 0
    indices[1] = 0
    indices[2] = 0
    indices[3] = 0
    indices[4] = 0
    pIndx = cast(pointer(indices), POINTER(agcom.LONG))
    lb = (agcom.LONG*5)()
    ub = (agcom.LONG*5)()
    for i in range(0,5):
        getlb = agcom.LONG()
        getub = agcom.LONG()
        agcom.OLEAut32Lib.SafeArrayGetLBound(sa, agcom.UINT(i+1), byref(getlb))
        agcom.OLEAut32Lib.SafeArrayGetUBound(sa, agcom.UINT(i+1), byref(getub))
        lb[i] = getlb
        ub[i] = getub
    for first in range(int(lb[0]), int(ub[0]) + 1):
        indices[0] = agcom.LONG(first)
        secondlvl = []
        for second in range(int(lb[1]), int(ub[1]) + 1):
            indices[1] = agcom.LONG(second)
            thirdlvl = []
            for third in range(int(lb[2]), int(ub[2]) + 1):
                indices[2]= agcom.LONG(third)
                fourthlvl = []
                for fourth in range(int(lb[3]), int(ub[3]) + 1):
                    indices[3] = agcom.LONG(fourth)
                    fifthlvl = []
                    for fifth in range(int(lb[4]), int(ub[4]) + 1):
                        indices[4] = agcom.LONG(fifth)
                        pElem = _vartype_to_ctypes_type(agcom.VT_VARIANT)()
                        agcom.OLEAut32Lib.VariantInit(pElem)
                        hr = agcom.OLEAut32Lib.SafeArrayGetElement(sa, pIndx, byref(pElem))
                        python_elem = python_val_from_ctypes_val(pElem, agcom.VT_VARIANT)
                        agcom.OLEAut32Lib.VariantClear(pElem)
                        fifthlvl.append(python_elem)
                    fourthlvl.append(fifthlvl)
                thirdlvl.append(fourthlvl)
            secondlvl.append(thirdlvl)
        python_array.append(secondlvl)
    return python_array

def list_from_SAFEARRAY(sa:agcom.LPSAFEARRAY) -> list:
    dim = agcom.OLEAut32Lib.SafeArrayGetDim(sa)
    if dim == 0:
        return list()
    elif dim == 1:
        return _single_dimension_list_from_SAFEARRAY(sa, 0)
    elif dim == 2:
        lb = agcom.LONG()
        ub = agcom.LONG()
        agcom.OLEAut32Lib.SafeArrayGetLBound(sa, agcom.UINT(1), byref(lb))
        agcom.OLEAut32Lib.SafeArrayGetUBound(sa, agcom.UINT(1), byref(ub))
        ret = []
        for i in range(int(lb.value), int(ub.value)+1):
            ret.append(_single_dimension_list_from_SAFEARRAY(sa, i, True))
        return ret
    elif dim == 5:
        return _five_dimension_list_from_SAFEARRAY(sa)
    else:
        raise RuntimeError("Unexpected dimension of SafeArray.  Expected 1 or 2, got " + str(dim) + ".")


###############################################################################
#   Classes for marshalling arguments to COM interfaces
###############################################################################
class LongArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.LONG = agcom.LONG()
        else:
            self.LONG = agcom.LONG(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.LONG:
        return self.LONG
    @property
    def python_val(self) -> int:
        return self.LONG.value

class ULongArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.ULONG = agcom.ULONG()
        else:
            self.ULONG = agcom.ULONG(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.ULONG:
        return self.ULONG
    @property
    def python_val(self) -> int:
        return self.ULONG.value

class LongLongArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.LONGLONG = agcom.LONGLONG()
        else:
            self.LONGLONG = agcom.LONGLONG(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.LONGLONG:
        return self.LONGLONG
    @property
    def python_val(self) -> int:
        return self.LONGLONG.value

class ULongLongArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.ULONGLONG = agcom.ULONGLONG()
        else:
            self.ULONGLONG = agcom.ULONGLONG(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.ULONGLONG:
        return self.ULONGLONG
    @property
    def python_val(self) -> int:
        return self.ULONGLONG.value

class IntArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.INT = agcom.INT()
        else:
            self.INT = agcom.INT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.INT:
        return self.INT
    @property
    def python_val(self) -> int:
        return self.INT.value

class UIntArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.UINT = agcom.UINT()
        else:
            self.UINT = agcom.UINT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.UINT:
        return self.UINT
    @property
    def python_val(self) -> int:
        return self.UINT.value

class ShortArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.SHORT = agcom.SHORT()
        else:
            self.SHORT = agcom.SHORT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.SHORT:
        return self.SHORT
    @property
    def python_val(self) -> int:
        return self.SHORT.value

class UShortArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.USHORT = agcom.USHORT()
        else:
            self.USHORT = agcom.USHORT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.USHORT:
        return self.USHORT
    @property
    def python_val(self) -> int:
        return self.USHORT.value

class HResultArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.HRESULT = agcom.HRESULT()
        else:
            self.HRESULT = agcom.HRESULT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.HRESULT:
        return self.HRESULT
    @property
    def python_val(self) -> int:
        return self.HRESULT.value

class OLEColorArg(object):
    def __init__(self, val: agcolor.Color = None):
        if val is None:
            self.OLE_COLOR = agcom.OLE_COLOR()
        else:
            if type(val) == agcolor.ColorRGBA:
                raise STKColorError("Argument type is RGB only, use Color class instead of ColorRGBA.")
            self.OLE_COLOR = agcom.OLE_COLOR(val._to_ole_color())
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.OLE_COLOR:
        return self.OLE_COLOR
    @property
    def python_val(self) -> agcolor.Color:
        c = agcolor.Color()
        c._from_ole_color(self.OLE_COLOR.value)
        return c

class VariantBoolArg(object):
    def __init__(self, val: bool = None):
        if val is None:
            self.vb = agcom.VARIANT_BOOL()
        else:
            self.vb = agcom.VARIANT_TRUE if val else agcom.VARIANT_FALSE
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.VARIANT_BOOL:
        return self.vb
    @property
    def python_val(self) -> bool:
        return False if self.vb.value == agcom.VARIANT_FALSE.value else True

class DoubleArg(object):
    def __init__(self, val: float = None):
        if val is None:
            self.DOUBLE = agcom.DOUBLE()
        else:
            self.DOUBLE = agcom.DOUBLE(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.DOUBLE:
        return self.DOUBLE
    @property
    def python_val(self) -> float:
        return self.DOUBLE.value

class FloatArg(object):
    def __init__(self, val: float = None):
        if val is None:
            self.FLOAT = agcom.FLOAT()
        else:
            self.FLOAT = agcom.FLOAT(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.FLOAT:
        return self.FLOAT
    @property
    def python_val(self) -> float:
        return self.FLOAT.value

class BStrArg(object):
    def __init__(self, val: str = None):
        if val is None:
            self.bstr = agcom.BSTR()
        else:
            self.bstr = agcom.BSTR(agcom.OLEAut32Lib.SysAllocString(val))
    def _cleanup(self):
        agcom.OLEAut32Lib.SysFreeString(self.bstr)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.BSTR:
        return self.bstr
    @property
    def python_val(self) -> str:
        if self.bstr.value is not None:
            return str(self.bstr.value)

class ByteArg(object):
    def __init__(self, val: typing.Any = None):
        if val is None:
            self.BYTE = agcom.BYTE()
        else:
            if type(val) == int:
                self.BYTE = agcom.BYTE(val)
            elif type(val) == str:
                self.BYTE = agcom.BYTE(ord(val[0]))
            elif type(val) == bytes:
                self.BYTE = agcom.BYTE(ord(val[0]))
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.BYTE:
        return self.BYTE
    @property
    def python_val(self) -> bytes:
        return bytes(chr(self.BYTE.value), "ascii")

class CharArg(object):
    def __init__(self, val: typing.Any = None):
        if val is None:
            self.CHAR = agcom.CHAR()
        else:
            if type(val) == int:
                self.CHAR = agcom.CHAR(val)
            elif type(val) == str:
                self.CHAR = agcom.CHAR(ord(val[0]))
            elif type(val) == bytes:
                self.CHAR = agcom.CHAR(ord(val[0]))
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.CHAR:
        return self.CHAR
    @property
    def python_val(self) -> bytes:
        return bytes(chr(self.CHAR.value), "ascii")

class VariantArg(object):
    def __init__(self, val: typing.Any = None):
        if val is not None and type(val)==agcom.Variant:
            self.var = agcom.Variant()
            agcom.OLEAut32Lib.VariantCopy(byref(self.var), byref(val))
        else:
            self.var = VARIANT_from_python_data(val)
    def _cleanup(self):
        agcom.OLEAut32Lib.VariantClear(self.var)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.Variant:
        return self.var
    @property
    def python_val(self) -> typing.Any:
        return python_val_from_VARIANT(self.var)

class InterfaceInArg(object):
    def __init__(self, as_interface):
        self.pIntf = None
        self.as_interface = as_interface
    def __call__(self, val):
        """
        Initialize an InterfaceInArg object.

        val should be a python CoClass object (e.g. Facility)
        as_interface is the interface class to send to STK
        """
        new_inst = InterfaceInArg(self.as_interface)
        new_inst.pIntf = None
        if type(val) == agcomobj.COMObject:
            new_inst.val = val
            new_inst.rawptr = val.get_pointer()
        elif val is not None and hasattr(val, "_intf"):
            new_inst.val = val
            if new_inst.as_interface=="IDispatch":
                new_inst.pIntf = val._intf.query_interface(agcom.GUID(agcom.IDispatch._guid))
            elif new_inst.as_interface=="IUnknown":
                new_inst.pIntf = val._intf.query_interface(agcom.GUID(agcom.IUnknown._guid))
            else:
                intf_class = agcoclass.AgTypeNameMap[new_inst.as_interface]
                new_inst.pIntf = val._intf.query_interface(intf_class._metadata)
            new_inst.rawptr = new_inst.pIntf.p
        else:
            new_inst.val = None
            new_inst.pIntf = agcom.IUnknown()
            new_inst.rawptr = new_inst.pIntf.p
        return new_inst
    def _cleanup(self):
        if self.pIntf is not None:
            del(self.pIntf)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        if hasattr(self.rawptr, "value"):
            return agcom.PVOID(self.rawptr.value)
        else:
            return agcom.PVOID(int(self.rawptr))
    @property
    def python_val(self) -> typing.Any:
        return self.val

class InterfaceOutArg(object):
    def __init__(self):
        self.ptr = agcom.IUnknown()
    def _cleanup(self):
        del(self.ptr)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        return self.ptr.p
    @property
    def python_val(self) -> typing.Any:
        if self.ptr:
            self.ptr.take_ownership()
            return agcoclass.get_concrete_class(self.ptr)
        else:
            return None

class PVoidArg(object):
    def __init__(self, val: agcom.PVOID = None):
        if val is None:
            self.p = agcom.PVOID()
        else:
            self.p = val
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        return self.p
    @property
    def python_val(self) -> agcom.PVOID:
        return self.p

class InterfaceEventCallbackArg(object):
    def __init__(self, pUnk:agcom.PVOID, as_interface):
        """
        Initialize an InterfaceEventCallbackArg object.

        pUnk should be a IUnknown pointer as PVOID
        as_interface is the interface class to send to STK
        """
        ptr = agcom.IUnknown()
        ptr.p = agcom.PVOID(pUnk) if type(pUnk)==int else pUnk
        self.intf = as_interface()
        self.intf._private_init(ptr)
        del(ptr)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        del(self.intf)
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        return self.intf.__dict__["_intf"].p
    @property
    def python_val(self) -> typing.Any:
        return self.intf

class IEnumVariantArg(object):
    def __init__(self):
        self.ptr = agcom.IUnknown()
    def _cleanup(self):
        del(self.ptr)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        return self.ptr.p
    @property
    def python_val(self) -> agcom.IEnumVariant:
        if self.ptr:
            self.ptr.take_ownership()
            return agcom.IEnumVariant(self.ptr)
        else:
            return None

class EnumArg(object):
    def __init__(self, enum_type: IntEnum):
        self.enum_type = enum_type
    def __call__(self, val: typing.Any = None):
        if val is None:
            self.val = agcom.LONG()
        elif type(val) == agcom.LONG:
            self.val = val
        else:
            self.val = agcom.LONG(val)
        return self
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.LONG:
        return self.val
    @property
    def python_val(self) -> IntEnum:
        return self.enum_type(self.val.value)

class LPSafearrayArg(object):
    def __init__(self, val: list = None):
        if val is None:
            self.sa = agcom.LPSAFEARRAY()
        else:
            self.sa = SAFEARRAY_from_list(val, True)
    def _cleanup(self):
        agcom.OLEAut32Lib.SafeArrayDestroy(self.sa)
    def __del__(self):
        self._cleanup()
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        self._cleanup()
        return False
    @property
    def com_val(self) -> agcom.LPSAFEARRAY:
        return self.sa
    @property
    def python_val(self) -> list:
        return list_from_SAFEARRAY(self.sa)

class IPictureDispArg(object):
    def __init__(self, val: agcom.IPictureDisp = None):
        raise STKPluginMethodNotImplementedError(f"Methods with the argument type \"IPictureDisp\" are not available using Python")
        if val is None:
            self.ipd = agcom.IPictureDisp()
        else:
            self.ipd = val
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.PVOID:
        return agcom.PVOID()
    @property
    def python_val(self) -> agcom.IPictureDisp:
        return self.ipd

class DateArg(object):
    def __init__(self, val: datetime = None):
        if val is None:
            self.date = agcom.DATE()
        else:
            self.date = datetime_to_DATE(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.DATE:
        return self.date
    @property
    def python_val(self) -> datetime:
        return DATE_to_datetime(self.date)

class OLEHandleArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.OLE_HANDLE = agcom.OLE_HANDLE()
        else:
            self.OLE_HANDLE = agcom.OLE_HANDLE(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.OLE_HANDLE:
        return self.OLE_HANDLE
    @property
    def python_val(self) -> int:
        return self.OLE_HANDLE.value

class OLEXPosPixelsArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.OLE_XPOS_PIXELS = agcom.OLE_XPOS_PIXELS()
        else:
            self.OLE_XPOS_PIXELS = agcom.OLE_XPOS_PIXELS(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.OLE_XPOS_PIXELS:
        return self.OLE_XPOS_PIXELS
    @property
    def python_val(self) -> int:
        return self.OLE_XPOS_PIXELS.value

class OLEYPosPixelsArg(object):
    def __init__(self, val: int = None):
        if val is None:
            self.OLE_YPOS_PIXELS = agcom.OLE_YPOS_PIXELS()
        else:
            self.OLE_YPOS_PIXELS = agcom.OLE_YPOS_PIXELS(val)
    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        return False
    @property
    def com_val(self) -> agcom.OLE_YPOS_PIXELS:
        return self.OLE_YPOS_PIXELS
    @property
    def python_val(self) -> int:
        return self.OLE_YPOS_PIXELS.value



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [ "COMEventHandlerImpl",
            "ISTKObjectRootEventCOMHandler",
            "ISTKXApplicationEventCOMHandler",
            "IGraphics2DControlEventCOMHandler",
            "IGraphics3DControlEventCOMHandler",
            "ISceneEventCOMHandler",
            "IKmlGraphicsEventCOMHandler",
            "IImageCollectionEventCOMHandler",
            "ITerrainOverlayCollectionEventCOMHandler"]

import os
from ctypes import CFUNCTYPE, POINTER, c_void_p, cast, addressof, Structure

from .                       import marshall     as agmarshall
from .                       import coclassutil  as agcls
from .comutil import (BSTR, DISPID, DispParams, DOUBLE, ExcepInfo, E_NOINTERFACE, E_NOTIMPL, GUID,
                      HRESULT, IDispatch, IUnknown, LCID, LONG, LPOLESTR, OLE_XPOS_PIXELS,
                      OLE_YPOS_PIXELS, PVOID, REFIID, SHORT, S_OK, UINT, ULONG, Variant, WORD,
                      pointer)

class COMEventHandlerImpl(object):
    _IID_IUnknown  = GUID.from_registry_format(IUnknown._guid)
    _IID_IDispatch = GUID.from_registry_format(IDispatch._guid)

    def __init__(self, pUnk, pUnkSink, iid):
        self._connection_id = None
        self._base_pUnkSink = pUnkSink
        self._cpc = agcls.IConnectionPointContainer(IUnknown(pUnk))
        self._cp = self._cpc.find_connection_point(iid)

    def __del__(self):
        del(self._cp)
        del(self._cpc)

    def _add_ref(self, pThis:PVOID) -> int:
        return 1

    def _release(self, pThis:PVOID) -> int:
        return 0

    def _get_type_info_count(self, pThis:PVOID, pctinfo:POINTER(UINT)) -> int:
        return E_NOTIMPL

    def _get_type_info(self, pThis:PVOID, iTInfo:UINT, lcid:LCID, ppTInfo:POINTER(PVOID)) -> int:
        return E_NOTIMPL

    def _get_ids_of_names(self, pThis:PVOID, riid:REFIID, rgszNames:POINTER(LPOLESTR), cNames:UINT, lcid:LCID, rgDispId:POINTER(DISPID)) -> int:
        return E_NOTIMPL

    def subscribe(self):
        if self._connection_id is None:
            self._connection_id = self._cp.advise(addressof(self._base_pUnkSink))

    def unsubscribe(self):
        if self._connection_id is not None:
            self._cp.unadvise(self._connection_id)
            self._connection_id = None

################################################################################
#          IAgStkObjectRootEvents
################################################################################

class _STKObjectRootRawEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",               c_void_p),
                 ("IUnknown2",               c_void_p),
                 ("IUnknown3",               c_void_p),
                 ("on_scenario_new",           c_void_p),
                 ("on_scenario_load",          c_void_p),
                 ("on_scenario_close",         c_void_p),
                 ("on_scenario_save",          c_void_p),
                 ("on_log_message",            c_void_p),
                 ("on_anim_update",            c_void_p),
                 ("on_stk_object_added",        c_void_p),
                 ("on_stk_object_deleted",      c_void_p),
                 ("on_stk_object_renamed",      c_void_p),
                 ("on_animation_playback",     c_void_p),
                 ("on_animation_rewind",       c_void_p),
                 ("on_animation_pause",        c_void_p),
                 ("on_scenario_before_save",    c_void_p),
                 ("on_animation_step",         c_void_p),
                 ("on_animation_step_back",     c_void_p),
                 ("on_animation_slower",       c_void_p),
                 ("on_animation_faster",       c_void_p),
                 ("on_percent_complete_update", c_void_p),
                 ("on_percent_complete_end",    c_void_p),
                 ("on_percent_complete_begin",  c_void_p),
                 ("on_stk_object_changed",      c_void_p),
                 ("on_scenario_before_close",   c_void_p),
                 ("on_stk_object_pre_delete",    c_void_p)]

class _STKObjectRootRawEvents2UnkSink(Structure):
    _fields_ = [ ("IUnknown1",                  c_void_p),
                 ("IUnknown2",                  c_void_p),
                 ("IUnknown3",                  c_void_p),
                 ("on_stk_object_start_3d_editing",  c_void_p),
                 ("on_stk_object_stop_3d_editing",   c_void_p),
                 ("on_stk_object_apply_3d_editing",  c_void_p),
                 ("on_stk_object_cancel_3d_editing", c_void_p),
                 ("on_stk_object_pre_cut",          c_void_p),
                 ("on_stk_object_copy",            c_void_p),
                 ("on_stk_object_paste",           c_void_p) ]

class ISTKObjectRootEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgStkObjectRootEvents     = GUID.from_registry_format("{4A25888C-BF0A-4B79-816B-2623D16042B0}")
    _IID_IAgStkObjectRootRawEvents  = GUID.from_registry_format("{A381FC71-ACBF-4034-B732-2A36B0CFA2E4}")
    _IID_IAgStkObjectRootRawEvents2 = GUID.from_registry_format("{F607E46E-A49F-4B9B-BE24-F29F63709FB0}")

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable1()
        self._init_vtable2()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink1, ISTKObjectRootEventCOMHandler._IID_IAgStkObjectRootEvents)

    def _init_vtable1(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]           = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]           = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]           = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]           = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]           = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]           = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_OnScenarioNew"]           = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_new)
        self.__dict__["_cfunc_OnScenarioLoad"]          = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_load)
        self.__dict__["_cfunc_OnScenarioClose"]         = CFUNCTYPE(None, PVOID)(self._on_scenario_close)
        self.__dict__["_cfunc_OnScenarioSave"]          = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_save)
        self.__dict__["_cfunc_OnLogMessage"]            = CFUNCTYPE(None, PVOID, BSTR, LONG, LONG, BSTR, LONG, LONG)(self._on_log_message)
        self.__dict__["_cfunc_OnAnimUpdate"]            = CFUNCTYPE(None, PVOID, DOUBLE)(self._on_anim_update)
        self.__dict__["_cfunc_OnStkObjectAdded"]        = CFUNCTYPE(None, PVOID, Variant)(self._on_stk_object_added)
        self.__dict__["_cfunc_OnStkObjectDeleted"]      = CFUNCTYPE(None, PVOID, Variant)(self._on_stk_object_deleted)
        self.__dict__["_cfunc_OnStkObjectRenamed"]      = CFUNCTYPE(None, PVOID, Variant, BSTR, BSTR)(self._on_stk_object_renamed)
        self.__dict__["_cfunc_OnAnimationPlayback"]     = CFUNCTYPE(None, PVOID, DOUBLE, LONG, LONG)(self._on_animation_playback)
        self.__dict__["_cfunc_OnAnimationRewind"]       = CFUNCTYPE(None, PVOID)(self._on_animation_rewind)
        self.__dict__["_cfunc_OnAnimationPause"]        = CFUNCTYPE(None, PVOID, DOUBLE)(self._on_animation_pause)
        self.__dict__["_cfunc_OnScenarioBeforeSave"]    = CFUNCTYPE(None, PVOID, PVOID)(self._on_scenario_before_save)
        self.__dict__["_cfunc_OnAnimationStep"]         = CFUNCTYPE(None, PVOID, DOUBLE)(self._on_animation_step)
        self.__dict__["_cfunc_OnAnimationStepBack"]     = CFUNCTYPE(None, PVOID, DOUBLE)(self._on_animation_step_back)
        self.__dict__["_cfunc_OnAnimationSlower"]       = CFUNCTYPE(None, PVOID)(self._on_animation_slower)
        self.__dict__["_cfunc_OnAnimationFaster"]       = CFUNCTYPE(None, PVOID)(self._on_animation_faster)
        self.__dict__["_cfunc_OnPercentCompleteUpdate"] = CFUNCTYPE(None, PVOID, PVOID)(self._on_percent_complete_update)
        self.__dict__["_cfunc_OnPercentCompleteEnd"]    = CFUNCTYPE(None, PVOID)(self._on_percent_complete_end)
        self.__dict__["_cfunc_OnPercentCompleteBegin"]  = CFUNCTYPE(None, PVOID)(self._on_percent_complete_begin)
        self.__dict__["_cfunc_OnStkObjectChanged"]      = CFUNCTYPE(None, PVOID, PVOID)(self._on_stk_object_changed)
        self.__dict__["_cfunc_OnScenarioBeforeClose"]   = CFUNCTYPE(None, PVOID)(self._on_scenario_before_close)
        self.__dict__["_cfunc_OnStkObjectPreDelete"]    = CFUNCTYPE(None, PVOID, PVOID)(self._on_stk_object_pre_delete)

        self.__dict__["_vtable1"] = _STKObjectRootRawEventsUnkSink( *[cast(self._cfunc_IUnknown1,               c_void_p),
                                                                        cast(self._cfunc_IUnknown2,               c_void_p),
                                                                        cast(self._cfunc_IUnknown3,               c_void_p),
                                                                        cast(self._cfunc_OnScenarioNew,           c_void_p),
                                                                        cast(self._cfunc_OnScenarioLoad,          c_void_p),
                                                                        cast(self._cfunc_OnScenarioClose,         c_void_p),
                                                                        cast(self._cfunc_OnScenarioSave,          c_void_p),
                                                                        cast(self._cfunc_OnLogMessage,            c_void_p),
                                                                        cast(self._cfunc_OnAnimUpdate,            c_void_p),
                                                                        cast(self._cfunc_OnStkObjectAdded,        c_void_p),
                                                                        cast(self._cfunc_OnStkObjectDeleted,      c_void_p),
                                                                        cast(self._cfunc_OnStkObjectRenamed,      c_void_p),
                                                                        cast(self._cfunc_OnAnimationPlayback,     c_void_p),
                                                                        cast(self._cfunc_OnAnimationRewind,       c_void_p),
                                                                        cast(self._cfunc_OnAnimationPause,        c_void_p),
                                                                        cast(self._cfunc_OnScenarioBeforeSave,    c_void_p),
                                                                        cast(self._cfunc_OnAnimationStep,         c_void_p),
                                                                        cast(self._cfunc_OnAnimationStepBack,     c_void_p),
                                                                        cast(self._cfunc_OnAnimationSlower,       c_void_p),
                                                                        cast(self._cfunc_OnAnimationFaster,       c_void_p),
                                                                        cast(self._cfunc_OnPercentCompleteUpdate, c_void_p),
                                                                        cast(self._cfunc_OnPercentCompleteEnd,    c_void_p),
                                                                        cast(self._cfunc_OnPercentCompleteBegin,  c_void_p),
                                                                        cast(self._cfunc_OnStkObjectChanged,      c_void_p),
                                                                        cast(self._cfunc_OnScenarioBeforeClose,   c_void_p),
                                                                        cast(self._cfunc_OnStkObjectPreDelete,    c_void_p)] )
        self.__dict__["_pUnkSink1"] = pointer(self._vtable1)

    def _init_vtable2(self):
        self.__dict__["_cfunc_OnStkObjectStart3dEditing"]  = CFUNCTYPE(None, PVOID, BSTR)(self._on_stk_object_start_3d_editing)
        self.__dict__["_cfunc_OnStkObjectStop3dEditing"]   = CFUNCTYPE(None, PVOID, BSTR)(self._on_stk_object_stop_3d_editing)
        self.__dict__["_cfunc_OnStkObjectApply3dEditing"]  = CFUNCTYPE(None, PVOID, BSTR)(self._on_stk_object_apply_3d_editing)
        self.__dict__["_cfunc_OnStkObjectCancel3dEditing"] = CFUNCTYPE(None, PVOID, BSTR)(self._on_stk_object_cancel_3d_editing)
        self.__dict__["_cfunc_OnStkObjectPreCut"]          = CFUNCTYPE(None, PVOID, PVOID)(self._on_stk_object_pre_cut)
        self.__dict__["_cfunc_OnStkObjectCopy"]            = CFUNCTYPE(None, PVOID, PVOID)(self._on_stk_object_copy)
        self.__dict__["_cfunc_OnStkObjectPaste"]           = CFUNCTYPE(None, PVOID, PVOID)(self._on_stk_object_paste)

        self.__dict__["_vtable2"] = _STKObjectRootRawEvents2UnkSink( *[cast(self._cfunc_IUnknown1,                  c_void_p),
                                                                         cast(self._cfunc_IUnknown2,                  c_void_p),
                                                                         cast(self._cfunc_IUnknown3,                  c_void_p),
                                                                         cast(self._cfunc_OnStkObjectStart3dEditing,  c_void_p),
                                                                         cast(self._cfunc_OnStkObjectStop3dEditing,   c_void_p),
                                                                         cast(self._cfunc_OnStkObjectApply3dEditing,  c_void_p),
                                                                         cast(self._cfunc_OnStkObjectCancel3dEditing, c_void_p),
                                                                         cast(self._cfunc_OnStkObjectPreCut,          c_void_p),
                                                                         cast(self._cfunc_OnStkObjectCopy,            c_void_p),
                                                                         cast(self._cfunc_OnStkObjectPaste,           c_void_p)] )
        self.__dict__["_pUnkSink2"] = pointer(self._vtable2)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = addressof(self._pUnkSink1)
            return S_OK
        elif iid == ISTKObjectRootEventCOMHandler._IID_IAgStkObjectRootEvents:
            ppvObject[0] = addressof(self._pUnkSink1)
            return S_OK
        elif iid == ISTKObjectRootEventCOMHandler._IID_IAgStkObjectRootRawEvents:
            ppvObject[0] = addressof(self._pUnkSink1)
            return S_OK
        elif iid == ISTKObjectRootEventCOMHandler._IID_IAgStkObjectRootRawEvents2:
            ppvObject[0] = addressof(self._pUnkSink2)
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _on_scenario_new(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioNew"]._callbacks:
            callback(path)

    def _on_scenario_load(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioLoad"]._callbacks:
            callback(path)

    def _on_scenario_close(self, pThis:PVOID) -> None:
        for callback in self._events["OnScenarioClose"]._callbacks:
            callback()

    def _on_scenario_save(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioSave"]._callbacks:
            callback(path)

    def _on_log_message(self, pThis:PVOID, message:str, msgType:int, errorCode:int, fileName:str, lineNo:int, dispID:int) -> None:
        for callback in self._events["OnLogMessage"]._callbacks:
            callback(message, agcls.AgTypeNameMap["LogMessageType"](msgType), errorCode, fileName, lineNo, agcls.AgTypeNameMap["LogMessageDisplayID"](dispID))

    def _on_anim_update(self, pThis:PVOID, timeEpSec:float) -> None:
        for callback in self._events["OnAnimUpdate"]._callbacks:
            callback(timeEpSec)

    def _on_stk_object_added(self, pThis:PVOID, Sender:Variant) -> None:
        for callback in self._events["OnStkObjectAdded"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender:
                callback(arg_Sender.python_val)

    def _on_stk_object_deleted(self, pThis:PVOID, Sender:Variant) -> None:
        for callback in self._events["OnStkObjectDeleted"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender:
                callback(arg_Sender.python_val)

    def _on_stk_object_renamed(self, pThis:PVOID, Sender:Variant, OldPath:str, NewPath:str) -> None:
        for callback in self._events["OnStkObjectRenamed"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender:
                callback(arg_Sender.python_val, OldPath, NewPath)

    def _on_animation_playback(self, pThis:PVOID, CurrentTime:float, eAction:int, eDirection:int) -> None:
        for callback in self._events["OnAnimationPlayback"]._callbacks:
            callback(CurrentTime, agcls.AgTypeNameMap["AnimationActionType"](eAction), agcls.AgTypeNameMap["AnimationDirectionType"](eDirection.python_val))

    def _on_animation_rewind(self, pThis:PVOID) -> None:
        for callback in self._events["OnAnimationRewind"]._callbacks:
            callback()

    def _on_animation_pause(self, pThis:PVOID, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationPause"]._callbacks:
            callback(CurrentTime)

    def _on_scenario_before_save(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnScenarioBeforeSave"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["ScenarioBeforeSaveEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_animation_step(self, pThis:PVOID, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationStep"]._callbacks:
            callback(CurrentTime)

    def _on_animation_step_back(self, pThis:PVOID, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationStepBack"]._callbacks:
            callback(CurrentTime)

    def _on_animation_slower(self, pThis:PVOID) -> None:
        for callback in self._events["OnAnimationSlower"]._callbacks:
            callback()

    def _on_animation_faster(self, pThis:PVOID) -> None:
        for callback in self._events["OnAnimationFaster"]._callbacks:
            callback()

    def _on_percent_complete_update(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnPercentCompleteUpdate"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["ProgressBarEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_percent_complete_end(self, pThis:PVOID) -> None:
        for callback in self._events["OnPercentCompleteEnd"]._callbacks:
            callback()

    def _on_percent_complete_begin(self, pThis:PVOID) -> None:
        for callback in self._events["OnPercentCompleteBegin"]._callbacks:
            callback()

    def _on_stk_object_changed(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnStkObjectChanged"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKObjectChangedEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_scenario_before_close(self, pThis:PVOID) -> None:
        for callback in self._events["OnScenarioBeforeClose"]._callbacks:
            callback()

    def _on_stk_object_pre_delete(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnStkObjectPreDelete"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKObjectPreDeleteEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_stk_object_start_3d_editing(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnStkObjectStart3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_stop_3d_editing(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnStkObjectStop3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_apply_3d_editing(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnStkObjectApply3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_cancel_3d_editing(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnStkObjectCancel3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_pre_cut(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnStkObjectPreCut"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKObjectCutCopyPasteEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_stk_object_copy(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnStkObjectCopy"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKObjectCutCopyPasteEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_stk_object_paste(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnStkObjectPaste"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKObjectCutCopyPasteEventArguments"]) as arg_pArgs:
                callback(arg_pArgs.python_val)


################################################################################
#          IAgSTKXApplicationEvents
################################################################################

class _STKXApplicationEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",                   c_void_p),
                 ("IUnknown2",                   c_void_p),
                 ("IUnknown3",                   c_void_p),
                 ("on_scenario_new",               c_void_p),
                 ("on_scenario_load",              c_void_p),
                 ("on_scenario_close",             c_void_p),
                 ("on_scenario_save",              c_void_p),
                 ("on_log_message",                c_void_p),
                 ("on_anim_update",                c_void_p),
                 ("on_new_globe_ctrl_request",       c_void_p),
                 ("on_new_map_ctrl_request",         c_void_p),
                 ("on_before_new_scenario",         c_void_p),
                 ("on_before_load_scenario",        c_void_p),
                 ("on_begin_scenario_close",        c_void_p),
                 ("on_new_gfx_analysis_ctrl_request", c_void_p),
                 ("on_ssl_certificate_server_error", c_void_p),
                 ("on_con_control_quit_received",    c_void_p) ]

class ISTKXApplicationEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgSTKXApplicationRawEvents = GUID.from_registry_format("{78C74BAF-7845-40BA-9EBE-C10FD081BC60}")
    _IID_IAgSTKXApplicationEvents    = GUID.from_registry_format("{3787DAB9-9A91-414B-B4EF-2339E0FBA96C}")

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, ISTKXApplicationEventCOMHandler._IID_IAgSTKXApplicationEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]               = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]               = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]               = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]               = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]               = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]               = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_OnScenarioNew"]               = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_new)
        self.__dict__["_cfunc_OnScenarioLoad"]              = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_load)
        self.__dict__["_cfunc_OnScenarioClose"]             = CFUNCTYPE(None, PVOID)(self._on_scenario_close)
        self.__dict__["_cfunc_OnScenarioSave"]              = CFUNCTYPE(None, PVOID, BSTR)(self._on_scenario_save)
        self.__dict__["_cfunc_OnLogMessage"]                = CFUNCTYPE(None, PVOID, BSTR, LONG, LONG, BSTR, LONG, LONG)(self._on_log_message)
        self.__dict__["_cfunc_OnAnimUpdate"]                = CFUNCTYPE(HRESULT, PVOID, DOUBLE)(self._on_anim_update)
        self.__dict__["_cfunc_OnNewGlobeCtrlRequest"]       = CFUNCTYPE(None, PVOID, LONG)(self._on_new_globe_ctrl_request)
        self.__dict__["_cfunc_OnNewMapCtrlRequest"]         = CFUNCTYPE(None, PVOID, LONG)(self._on_new_map_ctrl_request)
        self.__dict__["_cfunc_OnBeforeNewScenario"]         = CFUNCTYPE(None, PVOID, BSTR)(self._on_before_new_scenario)
        self.__dict__["_cfunc_OnBeforeLoadScenario"]        = CFUNCTYPE(None, PVOID, BSTR)(self._on_before_load_scenario)
        self.__dict__["_cfunc_OnBeginScenarioClose"]        = CFUNCTYPE(None, PVOID)(self._on_begin_scenario_close)
        self.__dict__["_cfunc_OnNewGfxAnalysisCtrlRequest"] = CFUNCTYPE(None, PVOID, LONG, LONG)(self._on_new_gfx_analysis_ctrl_request)
        self.__dict__["_cfunc_OnSSLCertificateServerError"] = CFUNCTYPE(None, PVOID, PVOID)(self._on_ssl_certificate_server_error)
        self.__dict__["_cfunc_OnConControlQuitReceived"]    = CFUNCTYPE(None, PVOID, PVOID)(self._on_con_control_quit_received)

        self.__dict__["_vtable"] = _STKXApplicationEventsUnkSink( *[cast(self._cfunc_IUnknown1,                   c_void_p),
                                                                      cast(self._cfunc_IUnknown2,                   c_void_p),
                                                                      cast(self._cfunc_IUnknown3,                   c_void_p),
                                                                      cast(self._cfunc_OnScenarioNew,               c_void_p),
                                                                      cast(self._cfunc_OnScenarioLoad,              c_void_p),
                                                                      cast(self._cfunc_OnScenarioClose,             c_void_p),
                                                                      cast(self._cfunc_OnScenarioSave,              c_void_p),
                                                                      cast(self._cfunc_OnLogMessage,                c_void_p),
                                                                      cast(self._cfunc_OnAnimUpdate,                c_void_p),
                                                                      cast(self._cfunc_OnNewGlobeCtrlRequest,       c_void_p),
                                                                      cast(self._cfunc_OnNewMapCtrlRequest,         c_void_p),
                                                                      cast(self._cfunc_OnBeforeNewScenario,         c_void_p),
                                                                      cast(self._cfunc_OnBeforeLoadScenario,        c_void_p),
                                                                      cast(self._cfunc_OnBeginScenarioClose,        c_void_p),
                                                                      cast(self._cfunc_OnNewGfxAnalysisCtrlRequest, c_void_p),
                                                                      cast(self._cfunc_OnSSLCertificateServerError, c_void_p),
                                                                      cast(self._cfunc_OnConControlQuitReceived,    c_void_p)] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        elif iid == ISTKXApplicationEventCOMHandler._IID_IAgSTKXApplicationRawEvents:
            ppvObject[0] = pThis
            return S_OK
        elif iid == ISTKXApplicationEventCOMHandler._IID_IAgSTKXApplicationEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _on_scenario_new(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioNew"]._callbacks:
            callback(path)

    def _on_scenario_load(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioLoad"]._callbacks:
            callback(path)

    def _on_scenario_close(self, pThis:PVOID) -> None:
        for callback in self._events["OnScenarioClose"]._callbacks:
            callback()

    def _on_scenario_save(self, pThis:PVOID, path:str) -> None:
        for callback in self._events["OnScenarioSave"]._callbacks:
            callback(path)

    def _on_log_message(self, pThis:PVOID, message:str, msgType:int, errorCode:int, fileName:str, lineNo:int, dispID:int) -> None:
        for callback in self._events["OnLogMessage"]._callbacks:
            callback(message, agcls.AgTypeNameMap["LogMessageType"](msgType), errorCode, fileName, lineNo, agcls.AgTypeNameMap["LogMessageDisplayID"](dispID))

    def _on_anim_update(self, pThis:PVOID, timeEpSec:float) -> int:
        for callback in self._events["OnAnimUpdate"]._callbacks:
            callback(timeEpSec)
        return S_OK

    def _on_new_globe_ctrl_request(self, pThis:PVOID, SceneID:int) -> None:
        for callback in self._events["OnNewGlobeCtrlRequest"]._callbacks:
            callback(SceneID)

    def _on_new_map_ctrl_request(self, pThis:PVOID, WinID:int) -> None:
        for callback in self._events["OnNewMapCtrlRequest"]._callbacks:
            callback(WinID)

    def _on_before_new_scenario(self, pThis:PVOID, Scenario:str) -> None:
        for callback in self._events["OnBeforeNewScenario"]._callbacks:
            callback(Scenario)

    def _on_before_load_scenario(self, pThis:PVOID, Scenario:str) -> None:
        for callback in self._events["OnBeforeLoadScenario"]._callbacks:
            callback(Scenario)

    def _on_begin_scenario_close(self, pThis:PVOID) -> None:
        for callback in self._events["OnBeginScenarioClose"]._callbacks:
            callback()

    def _on_new_gfx_analysis_ctrl_request(self, pThis:PVOID, SceneID:int, GfxAnalysisMode:int) -> None:
        for callback in self._events["OnNewGfxAnalysisCtrlRequest"]._callbacks:
            callback(SceneID, agcls.AgTypeNameMap["Graphics2DAnalysisMode"](GfxAnalysisMode))

    def _on_ssl_certificate_server_error(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnSSLCertificateServerError"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKXSSLCertificateErrorEventArgs"]) as arg_pArgs:
                callback(arg_pArgs.python_val)

    def _on_con_control_quit_received(self, pThis:PVOID, pArgs:PVOID) -> None:
        for callback in self._events["OnConControlQuitReceived"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(pArgs, agcls.AgTypeNameMap["STKXConControlQuitReceivedEventArgs"]) as arg_pArgs:
                callback(arg_pArgs.python_val)


################################################################################
#          ActiveX controls
################################################################################

class _UiAxStockEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",   c_void_p),
                 ("IUnknown2",   c_void_p),
                 ("IUnknown3",   c_void_p),
                 ("key_down",     c_void_p),
                 ("key_press",    c_void_p),
                 ("key_up",       c_void_p),
                 ("click",       c_void_p),
                 ("dbl_click",    c_void_p),
                 ("mouse_down",   c_void_p),
                 ("mouse_move",   c_void_p),
                 ("mouse_up",     c_void_p),
                 ("ole_drag_drop", c_void_p),
                 ("mouse_wheel",  c_void_p)]

class _Graphics3DControlEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",             c_void_p),
                 ("IUnknown2",             c_void_p),
                 ("IUnknown3",             c_void_p),
                 ("key_down",               c_void_p),
                 ("key_press",              c_void_p),
                 ("key_up",                 c_void_p),
                 ("click",                 c_void_p),
                 ("dbl_click",              c_void_p),
                 ("mouse_down",             c_void_p),
                 ("mouse_move",             c_void_p),
                 ("mouse_up",               c_void_p),
                 ("ole_drag_drop",           c_void_p),
                 ("mouse_wheel",            c_void_p),
                 ("on_object_editing_start",  c_void_p),
                 ("on_object_editing_apply",  c_void_p),
                 ("on_object_editing_cancel", c_void_p),
                 ("on_object_editing_stop",   c_void_p)]

class IUiAxStockEventCOMHandler(object):
    _IID_IAgUiAxStockRawEvents   = GUID.from_registry_format("{32A1F220-C90C-4FC6-B2BF-DF06DB89B72E}")

    def __init__(self, events:dict):
        self._events = events
        self.__dict__["_cfunc_KeyDown"]     = CFUNCTYPE(HRESULT, PVOID, POINTER(SHORT), SHORT)(self._key_down)
        self.__dict__["_cfunc_KeyPress"]    = CFUNCTYPE(HRESULT, PVOID, POINTER(SHORT))(self._key_press)
        self.__dict__["_cfunc_KeyUp"]       = CFUNCTYPE(HRESULT, PVOID, POINTER(SHORT), SHORT)(self._key_up)
        self.__dict__["_cfunc_Click"]       = CFUNCTYPE(HRESULT, PVOID)(self._click)
        self.__dict__["_cfunc_DblClick"]    = CFUNCTYPE(HRESULT, PVOID)(self._dbl_click)
        self.__dict__["_cfunc_MouseDown"]   = CFUNCTYPE(HRESULT, PVOID, SHORT, SHORT, OLE_XPOS_PIXELS, OLE_YPOS_PIXELS)(self._mouse_down)
        self.__dict__["_cfunc_MouseMove"]   = CFUNCTYPE(HRESULT, PVOID, SHORT, SHORT, OLE_XPOS_PIXELS, OLE_YPOS_PIXELS)(self._mouse_move)
        self.__dict__["_cfunc_MouseUp"]     = CFUNCTYPE(HRESULT, PVOID, SHORT, SHORT, OLE_XPOS_PIXELS, OLE_YPOS_PIXELS)(self._mouse_up)
        self.__dict__["_cfunc_OLEDragDrop"] = CFUNCTYPE(HRESULT, PVOID, PVOID, LONG, SHORT, SHORT, LONG, LONG)(self._ole_drag_drop)
        self.__dict__["_cfunc_MouseWheel"]  = CFUNCTYPE(HRESULT, PVOID, SHORT, SHORT, SHORT, OLE_XPOS_PIXELS, OLE_YPOS_PIXELS)(self._mouse_wheel)

    def _mouse_wheel(self, pThis:PVOID, Button:int, Shift:int, Delta:int, X:int, Y:int) -> int:
        for callback in self._events["MouseWheel"]._callbacks:
            callback(KeyCode, Shift, Delta, X, Y)
        return S_OK

    def _key_down(self, pThis:PVOID, KeyCode:POINTER(SHORT), Shift:int) -> int:
        for callback in self._events["KeyDown"]._callbacks:
            callback(KeyCode[0], Shift)
        return S_OK

    def _key_press(self, pThis:PVOID, KeyAscii:POINTER(SHORT)) -> int:
        for callback in self._events["KeyPress"]._callbacks:
            callback(KeyAscii[0])
        return S_OK

    def _key_up(self, pThis:PVOID, KeyCode:POINTER(SHORT), Shift:int) -> int:
        for callback in self._events["KeyUp"]._callbacks:
            callback(KeyCode[0], Shift)
        return S_OK

    def _click(self, pThis:PVOID) -> int:
        for callback in self._events["Click"]._callbacks:
            callback()
        return S_OK

    def _dbl_click(self, pThis:PVOID) -> int:
        for callback in self._events["DblClick"]._callbacks:
            callback()
        return S_OK

    def _mouse_down(self, pThis:PVOID, Button:int, Shift:int, X:int, Y:int) -> int:
        for callback in self._events["MouseDown"]._callbacks:
            callback(KeyCode, Shift, X, Y)
        return S_OK

    def _mouse_move(self, pThis:PVOID, Button:int, Shift:int, X:int, Y:int) -> int:
        for callback in self._events["MouseMove"]._callbacks:
            callback(KeyCode, Shift, X, Y)
        return S_OK

    def _mouse_up(self, pThis:PVOID, Button:int, Shift:int, X:int, Y:int) -> int:
        for callback in self._events["MouseUp"]._callbacks:
            callback(KeyCode, Shift, X, Y)
        return S_OK

    def _ole_drag_drop(self, pThis:PVOID, Data:PVOID, Effect:int, Button:int, Shift:int, X:int, Y:int) -> int:
        for callback in self._events["OLEDragDrop"]._callbacks:
            with agmarshall.InterfaceEventCallbackArg(Data, agcls.AgTypeNameMap["DataObject"]) as arg_Data:
                callback(arg_Data.python_val, Effect, KeyCode, Shift, X, Y)
        return S_OK

class IGraphics2DControlEventCOMHandler(COMEventHandlerImpl, IUiAxStockEventCOMHandler):
    _IID_IAgUiAx2DCntrlEvents    = GUID.from_registry_format("{DA0E1628-101E-4A18-B922-B4189E31AD7E}")

    def __init__(self, pUnk:IUnknown, events:dict):
        IUiAxStockEventCOMHandler.__init__(self, events)
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, IGraphics2DControlEventCOMHandler._IID_IAgUiAx2DCntrlEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"] = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"] = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"] = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"] = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"] = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"] = CFUNCTYPE(ULONG, PVOID)(self._release)

        self.__dict__["_vtable"] = _UiAxStockEventsUnkSink( *[cast(self._cfunc_IUnknown1,   c_void_p),
                                                                cast(self._cfunc_IUnknown2,   c_void_p),
                                                                cast(self._cfunc_IUnknown3,   c_void_p),
                                                                cast(self._cfunc_KeyDown,     c_void_p),
                                                                cast(self._cfunc_KeyPress,    c_void_p),
                                                                cast(self._cfunc_KeyUp,       c_void_p),
                                                                cast(self._cfunc_Click,       c_void_p),
                                                                cast(self._cfunc_DblClick,    c_void_p),
                                                                cast(self._cfunc_MouseDown,   c_void_p),
                                                                cast(self._cfunc_MouseMove,   c_void_p),
                                                                cast(self._cfunc_MouseUp,     c_void_p),
                                                                cast(self._cfunc_OLEDragDrop, c_void_p),
                                                                cast(self._cfunc_MouseWheel,  c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IUiAxStockEventCOMHandler._IID_IAgUiAxStockRawEvents:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IGraphics2DControlEventCOMHandler._IID_IAgUiAx2DCntrlEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

class IGraphics3DControlEventCOMHandler(COMEventHandlerImpl, IUiAxStockEventCOMHandler):
    _IID_IAgUiAxVOCntrlRawEvents = GUID.from_registry_format("{1ADE7AE0-B431-4ED4-8494-335EBB14007C}")
    _IID_IAgUiAxVOCntrlEvents    = GUID.from_registry_format("{C46F1BA0-22E4-432B-9259-C6DEF33FE2B2}")

    def __init__(self, pUnk:IUnknown, events:dict):
        IUiAxStockEventCOMHandler.__init__(self, events)
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, IGraphics3DControlEventCOMHandler._IID_IAgUiAxVOCntrlEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"] = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"] = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"] = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"] = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"] = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"] = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_OnObjectEditingStart"]  = CFUNCTYPE(HRESULT, PVOID, BSTR)(self._on_object_editing_start)
        self.__dict__["_cfunc_OnObjectEditingApply"]  = CFUNCTYPE(HRESULT, PVOID, BSTR)(self._on_object_editing_apply)
        self.__dict__["_cfunc_OnObjectEditingCancel"] = CFUNCTYPE(HRESULT, PVOID, BSTR)(self._on_object_editing_cancel)
        self.__dict__["_cfunc_OnObjectEditingStop"]   = CFUNCTYPE(HRESULT, PVOID, BSTR)(self._on_object_editing_stop)

        self.__dict__["_vtable"] = _Graphics3DControlEventsUnkSink( *[cast(self._cfunc_IUnknown1,             c_void_p),
                                                                  cast(self._cfunc_IUnknown2,             c_void_p),
                                                                  cast(self._cfunc_IUnknown3,             c_void_p),
                                                                  cast(self._cfunc_KeyDown,               c_void_p),
                                                                  cast(self._cfunc_KeyPress,              c_void_p),
                                                                  cast(self._cfunc_KeyUp,                 c_void_p),
                                                                  cast(self._cfunc_Click,                 c_void_p),
                                                                  cast(self._cfunc_DblClick,              c_void_p),
                                                                  cast(self._cfunc_MouseDown,             c_void_p),
                                                                  cast(self._cfunc_MouseMove,             c_void_p),
                                                                  cast(self._cfunc_MouseUp,               c_void_p),
                                                                  cast(self._cfunc_OLEDragDrop,           c_void_p),
                                                                  cast(self._cfunc_MouseWheel,            c_void_p),
                                                                  cast(self._cfunc_OnObjectEditingStart,  c_void_p),
                                                                  cast(self._cfunc_OnObjectEditingApply,  c_void_p),
                                                                  cast(self._cfunc_OnObjectEditingCancel, c_void_p),
                                                                  cast(self._cfunc_OnObjectEditingStop,   c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IUiAxStockEventCOMHandler._IID_IAgUiAxStockRawEvents:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IGraphics3DControlEventCOMHandler._IID_IAgUiAxVOCntrlRawEvents:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IGraphics3DControlEventCOMHandler._IID_IAgUiAxVOCntrlEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _on_object_editing_start(self, pThis:PVOID, Path:str) -> int:
        for callback in self._events["OnObjectEditingStart"]._callbacks:
            callback(Path)
        return S_OK

    def _on_object_editing_apply(self, pThis:PVOID, Path:str) -> int:
        for callback in self._events["OnObjectEditingApply"]._callbacks:
            callback(Path)
        return S_OK

    def _on_object_editing_cancel(self, pThis:PVOID, Path:str) -> int:
        for callback in self._events["OnObjectEditingCancel"]._callbacks:
            callback(Path)
        return S_OK

    def _on_object_editing_stop(self, pThis:PVOID, Path:str) -> int:
        for callback in self._events["OnObjectEditingStop"]._callbacks:
            callback(Path)
        return S_OK


################################################################################
#          IAgStkGraphicsSceneEvents
################################################################################

class _STKGraphicsSceneEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",        c_void_p),
                 ("IUnknown2",        c_void_p),
                 ("IUnknown3",        c_void_p),
                 ("GetTypeInfoCount", c_void_p),
                 ("GetTypeInfo",      c_void_p),
                 ("GetIDsOfNames",    c_void_p),
                 ("invoke",           c_void_p),
                 ("rendering",        c_void_p)]

class ISceneEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgStkGraphicsSceneEvents = GUID.from_registry_format("{FACA0112-848C-415D-B38C-0ED3F121D906}")
    _DISPID_Rendering = 13901

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, ISceneEventCOMHandler._IID_IAgStkGraphicsSceneEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_GetTypeInfoCount"] = CFUNCTYPE(HRESULT, PVOID, POINTER(UINT))(self._get_type_info_count)
        self.__dict__["_cfunc_GetTypeInfo"]      = CFUNCTYPE(HRESULT, PVOID, UINT, LCID, POINTER(PVOID))(self._get_type_info)
        self.__dict__["_cfunc_GetIDsOfNames"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(LPOLESTR), UINT, LCID, POINTER(DISPID))(self._get_ids_of_names)
        self.__dict__["_cfunc_Invoke"]           = CFUNCTYPE(HRESULT, PVOID, DISPID, REFIID, LCID, WORD, POINTER(DispParams), POINTER(Variant), POINTER(ExcepInfo), POINTER(UINT))(self._invoke)
        self.__dict__["_cfunc_Rendering"]        = CFUNCTYPE(HRESULT, PVOID, Variant, PVOID)(self._rendering)

        self.__dict__["_vtable"] = _STKGraphicsSceneEventsUnkSink( *[cast(self._cfunc_IUnknown1,        c_void_p),
                                                                       cast(self._cfunc_IUnknown2,        c_void_p),
                                                                       cast(self._cfunc_IUnknown3,        c_void_p),
                                                                       cast(self._cfunc_GetTypeInfoCount, c_void_p),
                                                                       cast(self._cfunc_GetTypeInfo,      c_void_p),
                                                                       cast(self._cfunc_GetIDsOfNames,    c_void_p),
                                                                       cast(self._cfunc_Invoke,           c_void_p),
                                                                       cast(self._cfunc_Rendering,        c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        if iid == COMEventHandlerImpl._IID_IDispatch:
            ppvObject[0] = pThis
            return S_OK
        elif iid == ISceneEventCOMHandler._IID_IAgStkGraphicsSceneEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _invoke(self, pThis:PVOID, dispIdMember:DISPID, riid:REFIID, lcid:LCID, wFlags:WORD, pDispParams:POINTER(DispParams), pVarResult:POINTER(Variant), pExcepInfo:POINTER(ExcepInfo), puArgErr:POINTER(UINT)) -> int:
        if dispIdMember == ISceneEventCOMHandler._DISPID_Rendering:
            variant_Sender = pDispParams.contents.rgvarg[1]
            pArgs = agmarshall.ctype_val_from_VARIANT(pDispParams.contents.rgvarg[0])
            self._rendering(pThis, variant_Sender, pArgs)
            return S_OK
        else:
            return E_NOINTERFACE

    def _rendering(self, pThis:PVOID, Sender:Variant, Args:PVOID) -> None:
        for callback in self._events["Rendering"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender, \
                 agmarshall.InterfaceEventCallbackArg(Args, agcls.AgTypeNameMap["RenderingEventArgs"]) as arg_Args:
                callback(arg_Sender.python_val, arg_Args.python_val)


################################################################################
#          IAgStkGraphicsKmlGraphicsEvents
################################################################################

class _STKGraphicsKmlGraphicsEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",        c_void_p),
                 ("IUnknown2",        c_void_p),
                 ("IUnknown3",        c_void_p),
                 ("GetTypeInfoCount", c_void_p),
                 ("GetTypeInfo",      c_void_p),
                 ("GetIDsOfNames",    c_void_p),
                 ("invoke",           c_void_p),
                 ("document_loaded",   c_void_p)]

class IKmlGraphicsEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgStkGraphicsKmlGraphicsEvents = GUID.from_registry_format("{0B64622D-307A-4549-9692-7F15F4D9AC94}")
    _DISPID_DocumentLoaded = 27101

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, IKmlGraphicsEventCOMHandler._IID_IAgStkGraphicsKmlGraphicsEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_GetTypeInfoCount"] = CFUNCTYPE(HRESULT, PVOID, POINTER(UINT))(self._get_type_info_count)
        self.__dict__["_cfunc_GetTypeInfo"]      = CFUNCTYPE(HRESULT, PVOID, UINT, LCID, POINTER(PVOID))(self._get_type_info)
        self.__dict__["_cfunc_GetIDsOfNames"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(LPOLESTR), UINT, LCID, POINTER(DISPID))(self._get_ids_of_names)
        self.__dict__["_cfunc_Invoke"]           = CFUNCTYPE(HRESULT, PVOID, DISPID, REFIID, LCID, WORD, POINTER(DispParams), POINTER(Variant), POINTER(ExcepInfo), POINTER(UINT))(self._invoke)
        self.__dict__["_cfunc_DocumentLoaded"]   = CFUNCTYPE(HRESULT, PVOID, Variant, PVOID)(self._document_loaded)

        self.__dict__["_vtable"] = _STKGraphicsKmlGraphicsEventsUnkSink( *[cast(self._cfunc_IUnknown1,        c_void_p),
                                                                             cast(self._cfunc_IUnknown2,        c_void_p),
                                                                             cast(self._cfunc_IUnknown3,        c_void_p),
                                                                             cast(self._cfunc_GetTypeInfoCount, c_void_p),
                                                                             cast(self._cfunc_GetTypeInfo,      c_void_p),
                                                                             cast(self._cfunc_GetIDsOfNames,    c_void_p),
                                                                             cast(self._cfunc_Invoke,           c_void_p),
                                                                             cast(self._cfunc_DocumentLoaded,   c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        if iid == COMEventHandlerImpl._IID_IDispatch:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IKmlGraphicsEventCOMHandler._IID_IAgStkGraphicsKmlGraphicsEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _invoke(self, pThis:PVOID, dispIdMember:DISPID, riid:REFIID, lcid:LCID, wFlags:WORD, pDispParams:POINTER(DispParams), pVarResult:POINTER(Variant), pExcepInfo:POINTER(ExcepInfo), puArgErr:POINTER(UINT)) -> int:
        if dispIdMember == IKmlGraphicsEventCOMHandler._DISPID_DocumentLoaded:
            variant_Sender = pDispParams.contents.rgvarg[1]
            pArgs = agmarshall.ctype_val_from_VARIANT(pDispParams.contents.rgvarg[0])
            self._document_loaded(pThis, variant_Sender, pArgs)
            return S_OK
        else:
            return E_NOINTERFACE

    def _document_loaded(self, pThis:PVOID, Sender:Variant, Args:PVOID) -> None:
        for callback in self._events["DocumentLoaded"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender, \
                 agmarshall.InterfaceEventCallbackArg(Args, agcls.AgTypeNameMap["KmlDocumentLoadedEventArgs"]) as arg_Args:
                callback(arg_Sender.python_val, arg_Args.python_val)


################################################################################
#          IAgStkGraphicsImageCollectionEvents
################################################################################

class _STKGraphicsImageCollectionEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",        c_void_p),
                 ("IUnknown2",        c_void_p),
                 ("IUnknown3",        c_void_p),
                 ("GetTypeInfoCount", c_void_p),
                 ("GetTypeInfo",      c_void_p),
                 ("GetIDsOfNames",    c_void_p),
                 ("invoke",           c_void_p),
                 ("add_complete",      c_void_p)]

class IImageCollectionEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgStkGraphicsImageCollectionEvents = GUID.from_registry_format("{150DFFDA-DD8C-4227-9B7D-F813277BCB8E}")
    _DISPID_AddComplete = 13301

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, IImageCollectionEventCOMHandler._IID_IAgStkGraphicsImageCollectionEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_GetTypeInfoCount"] = CFUNCTYPE(HRESULT, PVOID, POINTER(UINT))(self._get_type_info_count)
        self.__dict__["_cfunc_GetTypeInfo"]      = CFUNCTYPE(HRESULT, PVOID, UINT, LCID, POINTER(PVOID))(self._get_type_info)
        self.__dict__["_cfunc_GetIDsOfNames"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(LPOLESTR), UINT, LCID, POINTER(DISPID))(self._get_ids_of_names)
        self.__dict__["_cfunc_Invoke"]           = CFUNCTYPE(HRESULT, PVOID, DISPID, REFIID, LCID, WORD, POINTER(DispParams), POINTER(Variant), POINTER(ExcepInfo), POINTER(UINT))(self._invoke)
        self.__dict__["_cfunc_AddComplete"]      = CFUNCTYPE(HRESULT, PVOID, Variant, PVOID)(self._add_complete)

        self.__dict__["_vtable"] = _STKGraphicsImageCollectionEventsUnkSink( *[cast(self._cfunc_IUnknown1,        c_void_p),
                                                                                 cast(self._cfunc_IUnknown2,        c_void_p),
                                                                                 cast(self._cfunc_IUnknown3,        c_void_p),
                                                                                 cast(self._cfunc_GetTypeInfoCount, c_void_p),
                                                                                 cast(self._cfunc_GetTypeInfo,      c_void_p),
                                                                                 cast(self._cfunc_GetIDsOfNames,    c_void_p),
                                                                                 cast(self._cfunc_Invoke,           c_void_p),
                                                                                 cast(self._cfunc_AddComplete,      c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        if iid == COMEventHandlerImpl._IID_IDispatch:
            ppvObject[0] = pThis
            return S_OK
        elif iid == IImageCollectionEventCOMHandler._IID_IAgStkGraphicsImageCollectionEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _invoke(self, pThis:PVOID, dispIdMember:DISPID, riid:REFIID, lcid:LCID, wFlags:WORD, pDispParams:POINTER(DispParams), pVarResult:POINTER(Variant), pExcepInfo:POINTER(ExcepInfo), puArgErr:POINTER(UINT)) -> int:
        if dispIdMember == IImageCollectionEventCOMHandler._DISPID_AddComplete:
            variant_Sender = pDispParams.contents.rgvarg[1]
            pArgs = agmarshall.ctype_val_from_VARIANT(pDispParams.contents.rgvarg[0])
            self._add_complete(pThis, variant_Sender, pArgs)
            return S_OK
        else:
            return E_NOINTERFACE

    def _add_complete(self, pThis:PVOID, Sender:Variant, Args:PVOID) -> None:
        for callback in self._events["AddComplete"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender, \
                 agmarshall.InterfaceEventCallbackArg(Args, agcls.AgTypeNameMap["GlobeImageOverlayAddCompleteEventArgs"]) as arg_Args:
                callback(arg_Sender.python_val, arg_Args.python_val)


################################################################################
#          IAgStkGraphicsTerrainCollectionEvents
################################################################################

class _STKGraphicsTerrainCollectionEventsUnkSink(Structure):
    _fields_ = [ ("IUnknown1",        c_void_p),
                 ("IUnknown2",        c_void_p),
                 ("IUnknown3",        c_void_p),
                 ("GetTypeInfoCount", c_void_p),
                 ("GetTypeInfo",      c_void_p),
                 ("GetIDsOfNames",    c_void_p),
                 ("invoke",           c_void_p),
                 ("add_complete",      c_void_p)]

class ITerrainOverlayCollectionEventCOMHandler(COMEventHandlerImpl):
    _IID_IAgStkGraphicsTerrainCollectionEvents = GUID.from_registry_format("{0744D80B-C88B-4F1D-BF3F-78A5C3AB69BA}")
    _DISPID_AddComplete = 13401

    def __init__(self, pUnk:IUnknown, events:dict):
        self._events = events
        self._init_vtable()
        COMEventHandlerImpl.__init__(self, pUnk, self._pUnkSink, ITerrainOverlayCollectionEventCOMHandler._IID_IAgStkGraphicsTerrainCollectionEvents)

    def _init_vtable(self):
        if os.name == "nt":
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        else:
            self.__dict__["_cfunc_IUnknown3"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
            self.__dict__["_cfunc_IUnknown1"]    = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
            self.__dict__["_cfunc_IUnknown2"]    = CFUNCTYPE(ULONG, PVOID)(self._release)
        self.__dict__["_cfunc_GetTypeInfoCount"] = CFUNCTYPE(HRESULT, PVOID, POINTER(UINT))(self._get_type_info_count)
        self.__dict__["_cfunc_GetTypeInfo"]      = CFUNCTYPE(HRESULT, PVOID, UINT, LCID, POINTER(PVOID))(self._get_type_info)
        self.__dict__["_cfunc_GetIDsOfNames"]    = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(LPOLESTR), UINT, LCID, POINTER(DISPID))(self._get_ids_of_names)
        self.__dict__["_cfunc_Invoke"]           = CFUNCTYPE(HRESULT, PVOID, DISPID, REFIID, LCID, WORD, POINTER(DispParams), POINTER(Variant), POINTER(ExcepInfo), POINTER(UINT))(self._invoke)
        self.__dict__["_cfunc_AddComplete"]      = CFUNCTYPE(HRESULT, PVOID, Variant, PVOID)(self._add_complete)

        self.__dict__["_vtable"] = _STKGraphicsTerrainCollectionEventsUnkSink( *[cast(self._cfunc_IUnknown1,        c_void_p),
                                                                                   cast(self._cfunc_IUnknown2,        c_void_p),
                                                                                   cast(self._cfunc_IUnknown3,        c_void_p),
                                                                                   cast(self._cfunc_GetTypeInfoCount, c_void_p),
                                                                                   cast(self._cfunc_GetTypeInfo,      c_void_p),
                                                                                   cast(self._cfunc_GetIDsOfNames,    c_void_p),
                                                                                   cast(self._cfunc_Invoke,           c_void_p),
                                                                                   cast(self._cfunc_AddComplete,      c_void_p) ] )
        self.__dict__["_pUnkSink"] = pointer(self._vtable)

    def _query_interface(self, pThis:PVOID, riid:REFIID, ppvObject:POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == COMEventHandlerImpl._IID_IUnknown:
            ppvObject[0] = pThis
            return S_OK
        if iid == COMEventHandlerImpl._IID_IDispatch:
            ppvObject[0] = pThis
            return S_OK
        elif iid == ITerrainOverlayCollectionEventCOMHandler._IID_IAgStkGraphicsTerrainCollectionEvents:
            ppvObject[0] = pThis
            return S_OK
        else:
            ppvObject[0] = 0
            return E_NOINTERFACE

    def _invoke(self, pThis:PVOID, dispIdMember:DISPID, riid:REFIID, lcid:LCID, wFlags:WORD, pDispParams:POINTER(DispParams), pVarResult:POINTER(Variant), pExcepInfo:POINTER(ExcepInfo), puArgErr:POINTER(UINT)) -> int:
        if dispIdMember == ITerrainOverlayCollectionEventCOMHandler._DISPID_AddComplete:
            variant_Sender = pDispParams.contents.rgvarg[1]
            pArgs = agmarshall.ctype_val_from_VARIANT(pDispParams.contents.rgvarg[0])
            self._add_complete(pThis, variant_Sender, pArgs)
            return S_OK
        else:
            return E_NOINTERFACE

    def _add_complete(self, pThis:PVOID, Sender:Variant, Args:PVOID) -> None:
        for callback in self._events["AddComplete"]._callbacks:
            with agmarshall.VariantArg(Sender) as arg_Sender, \
                 agmarshall.InterfaceEventCallbackArg(Args, agcls.AgTypeNameMap["TerrainOverlayAddCompleteEventArgs"]) as arg_Args:
                callback(arg_Sender.python_val, arg_Args.python_val)


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [ "GrpcEventHandlerImpl",
            "ISTKObjectRootEventGrpcHandler",
            "ISTKXApplicationEventGrpcHandler",
            "ISceneEventGrpcHandler",
            "IKmlGraphicsEventGrpcHandler",
            "IImageCollectionEventGrpcHandler",
            "ITerrainOverlayCollectionEventGrpcHandler"]

import typing

from .grpcutil import GrpcInterface
from .coclassutil import AgTypeNameMap
from . import AgGrpcServices_pb2

class _iadd_callback(object):
    def __init__(self, intf, event_name, handler, callback_func):
        self._intf = intf
        self._event_name = event_name
        self._handler = handler
        self._func = callback_func

    def __call__(self, count):
        if count == 1:
            self._intf.subscribe(self._handler, self._event_name, self._func)

class _isub_callback(object):
    def __init__(self, intf, event_name, handler, callback_func):
        self._intf = intf
        self._event_name = event_name
        self._handler = handler
        self._func = callback_func

    def __call__(self, count):
        if count == 0:
            self._intf.unsubscribe(self._handler, self._event_name, self._func)

class GrpcEventHandlerImpl(object):
    def __init__(self, interface:GrpcInterface, event_handler:AgGrpcServices_pb2.EventHandler, events:dict):
        self._intf = interface
        self._handler = event_handler
        self._events = events

    def __del__(self):
        pass

    def _register_iadd_isub_callbacks(self, event:str, callback):
        self._events[event]._set_iadd_callback(_iadd_callback(self._intf, event, self._handler, callback))
        self._events[event]._set_isub_callback(_isub_callback(self._intf, event, self._handler, callback))

    def subscribe(self):
        # No action needed since events are subscribed individually rather than all at once
        pass

    def unsubscribe(self):
        self._intf.unsubscribe_all(self._handler)

################################################################################
#          IStkObjectRootEvents
################################################################################

class ISTKObjectRootEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgStkObjectRootEvents, events)
        self._register_iadd_isub_callbacks("OnScenarioNew", self._on_scenario_new)
        self._register_iadd_isub_callbacks("OnScenarioLoad", self._on_scenario_load)
        self._register_iadd_isub_callbacks("OnScenarioClose", self._on_scenario_close)
        self._register_iadd_isub_callbacks("OnScenarioSave", self._on_scenario_save)
        self._register_iadd_isub_callbacks("OnLogMessage", self._on_log_message)
        self._register_iadd_isub_callbacks("OnAnimUpdate", self._on_anim_update)
        self._register_iadd_isub_callbacks("OnStkObjectAdded", self._on_stk_object_added)
        self._register_iadd_isub_callbacks("OnStkObjectDeleted", self._on_stk_object_deleted)
        self._register_iadd_isub_callbacks("OnStkObjectRenamed", self._on_stk_object_renamed)
        self._register_iadd_isub_callbacks("OnAnimationPlayback", self._on_animation_playback)
        self._register_iadd_isub_callbacks("OnAnimationRewind", self._on_animation_rewind)
        self._register_iadd_isub_callbacks("OnAnimationPause", self._on_animation_pause)
        self._register_iadd_isub_callbacks("OnScenarioBeforeSave", self._on_scenario_before_save)
        self._register_iadd_isub_callbacks("OnAnimationStep", self._on_animation_step)
        self._register_iadd_isub_callbacks("OnAnimationStepBack", self._on_animation_step_back)
        self._register_iadd_isub_callbacks("OnAnimationSlower", self._on_animation_slower)
        self._register_iadd_isub_callbacks("OnAnimationFaster", self._on_animation_faster)
        self._register_iadd_isub_callbacks("OnPercentCompleteUpdate", self._on_percent_complete_update)
        self._register_iadd_isub_callbacks("OnPercentCompleteEnd", self._on_percent_complete_end)
        self._register_iadd_isub_callbacks("OnPercentCompleteBegin", self._on_percent_complete_begin)
        self._register_iadd_isub_callbacks("OnStkObjectChanged", self._on_stk_object_changed)
        self._register_iadd_isub_callbacks("OnScenarioBeforeClose", self._on_scenario_before_close)
        self._register_iadd_isub_callbacks("OnStkObjectPreDelete", self._on_stk_object_pre_delete)
        self._register_iadd_isub_callbacks("OnStkObjectStart3dEditing", self._on_stk_object_start_3d_editing)
        self._register_iadd_isub_callbacks("OnStkObjectStop3dEditing", self._on_stk_object_stop_3d_editing)
        self._register_iadd_isub_callbacks("OnStkObjectApply3dEditing", self._on_stk_object_apply_3d_editing)
        self._register_iadd_isub_callbacks("OnStkObjectCancel3dEditing", self._on_stk_object_cancel_3d_editing)
        self._register_iadd_isub_callbacks("OnStkObjectPreCut", self._on_stk_object_pre_cut)
        self._register_iadd_isub_callbacks("OnStkObjectCopy", self._on_stk_object_copy)
        self._register_iadd_isub_callbacks("OnStkObjectPaste", self._on_stk_object_paste)

    def _on_scenario_new(self, path:str) -> None:
        for callback in self._events["OnScenarioNew"]._callbacks:
            callback(path)

    def _on_scenario_load(self, path:str) -> None:
        for callback in self._events["OnScenarioLoad"]._callbacks:
            callback(path)

    def _on_scenario_close(self) -> None:
        for callback in self._events["OnScenarioClose"]._callbacks:
            callback()

    def _on_scenario_save(self, path:str) -> None:
        for callback in self._events["OnScenarioSave"]._callbacks:
            callback(path)

    def _on_log_message(self, message:str, msgType:int, errorCode:int, fileName:str, lineNo:int, dispID:int) -> None:
        for callback in self._events["OnLogMessage"]._callbacks:
            callback(message, AgTypeNameMap["LogMessageType"](msgType), errorCode, fileName, lineNo, AgTypeNameMap["LogMessageDisplayID"](dispID))

    def _on_anim_update(self, timeEpSec:float) -> None:
        for callback in self._events["OnAnimUpdate"]._callbacks:
            callback(timeEpSec)

    def _on_stk_object_added(self, Sender:typing.Any) -> None:
        for callback in self._events["OnStkObjectAdded"]._callbacks:
            callback(Sender)

    def _on_stk_object_deleted(self, Sender:typing.Any) -> None:
        for callback in self._events["OnStkObjectDeleted"]._callbacks:
            callback(Sender)

    def _on_stk_object_renamed(self, Sender:typing.Any, OldPath:str, NewPath:str) -> None:
        for callback in self._events["OnStkObjectRenamed"]._callbacks:
            callback(Sender, OldPath, NewPath)

    def _on_animation_playback(self, CurrentTime:float, eAction:int, eDirection:int) -> None:
        for callback in self._events["OnAnimationPlayback"]._callbacks:
            callback(CurrentTime, AgTypeNameMap["AnimationActionType"](eAction), AgTypeNameMap["AnimationDirectionType"](eDirection))

    def _on_animation_rewind(self) -> None:
        for callback in self._events["OnAnimationRewind"]._callbacks:
            callback()

    def _on_animation_pause(self, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationPause"]._callbacks:
            callback(CurrentTime)

    def _on_scenario_before_save(self, pArgs:"ScenarioBeforeSaveEventArguments") -> None:
        for callback in self._events["OnScenarioBeforeSave"]._callbacks:
            callback(pArgs)

    def _on_animation_step(self, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationStep"]._callbacks:
            callback(CurrentTime)

    def _on_animation_step_back(self, CurrentTime:float) -> None:
        for callback in self._events["OnAnimationStepBack"]._callbacks:
            callback(CurrentTime)

    def _on_animation_slower(self) -> None:
        for callback in self._events["OnAnimationSlower"]._callbacks:
            callback()

    def _on_animation_faster(self) -> None:
        for callback in self._events["OnAnimationFaster"]._callbacks:
            callback()

    def _on_percent_complete_update(self, pArgs:"ProgressBarEventArguments") -> None:
        for callback in self._events["OnPercentCompleteUpdate"]._callbacks:
            callback(pArgs)

    def _on_percent_complete_end(self) -> None:
        for callback in self._events["OnPercentCompleteEnd"]._callbacks:
            callback()

    def _on_percent_complete_begin(self) -> None:
        for callback in self._events["OnPercentCompleteBegin"]._callbacks:
            callback()

    def _on_stk_object_changed(self, pArgs:"STKObjectChangedEventArguments") -> None:
        for callback in self._events["OnStkObjectChanged"]._callbacks:
            callback(pArgs)

    def _on_scenario_before_close(self) -> None:
        for callback in self._events["OnScenarioBeforeClose"]._callbacks:
            callback()

    def _on_stk_object_pre_delete(self, pArgs:"STKObjectPreDeleteEventArguments") -> None:
        for callback in self._events["OnStkObjectPreDelete"]._callbacks:
            callback(pArgs)

    def _on_stk_object_start_3d_editing(self, path:str) -> None:
        for callback in self._events["OnStkObjectStart3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_stop_3d_editing(self, path:str) -> None:
        for callback in self._events["OnStkObjectStop3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_apply_3d_editing(self, path:str) -> None:
        for callback in self._events["OnStkObjectApply3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_cancel_3d_editing(self, path:str) -> None:
        for callback in self._events["OnStkObjectCancel3dEditing"]._callbacks:
            callback(path)

    def _on_stk_object_pre_cut(self, pArgs:"STKObjectCutCopyPasteEventArguments") -> None:
        for callback in self._events["OnStkObjectPreCut"]._callbacks:
            callback(pArgs)

    def _on_stk_object_copy(self, pArgs:"STKObjectCutCopyPasteEventArguments") -> None:
        for callback in self._events["OnStkObjectCopy"]._callbacks:
            callback(pArgs)

    def _on_stk_object_paste(self, pArgs:"STKObjectCutCopyPasteEventArguments") -> None:
        for callback in self._events["OnStkObjectPaste"]._callbacks:
            callback(pArgs)


################################################################################
#          ISTKXApplicationEvents
################################################################################

class ISTKXApplicationEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgSTKXApplicationEvents, events)
        self._register_iadd_isub_callbacks("OnScenarioNew", self._on_scenario_new)
        self._register_iadd_isub_callbacks("OnScenarioLoad", self._on_scenario_load)
        self._register_iadd_isub_callbacks("OnScenarioClose", self._on_scenario_close)
        self._register_iadd_isub_callbacks("OnScenarioSave", self._on_scenario_save)
        self._register_iadd_isub_callbacks("OnLogMessage", self._on_log_message)
        self._register_iadd_isub_callbacks("OnAnimUpdate", self._on_anim_update)
        self._register_iadd_isub_callbacks("OnNewGlobeCtrlRequest", self._on_new_globe_ctrl_request)
        self._register_iadd_isub_callbacks("OnNewMapCtrlRequest", self._on_new_map_ctrl_request)
        self._register_iadd_isub_callbacks("OnBeforeNewScenario", self._on_before_new_scenario)
        self._register_iadd_isub_callbacks("OnBeforeLoadScenario", self._on_before_load_scenario)
        self._register_iadd_isub_callbacks("OnBeginScenarioClose", self._on_begin_scenario_close)
        self._register_iadd_isub_callbacks("OnNewGfxAnalysisCtrlRequest", self._on_new_gfx_analysis_ctrl_request)
        self._register_iadd_isub_callbacks("OnSSLCertificateServerError", self._on_ssl_certificate_server_error)
        self._register_iadd_isub_callbacks("OnConControlQuitReceived", self._on_con_control_quit_received)

    def _on_scenario_new(self, path:str) -> None:
        for callback in self._events["OnScenarioNew"]._callbacks:
            callback(path)

    def _on_scenario_load(self, path:str) -> None:
        for callback in self._events["OnScenarioLoad"]._callbacks:
            callback(path)

    def _on_scenario_close(self) -> None:
        for callback in self._events["OnScenarioClose"]._callbacks:
            callback()

    def _on_scenario_save(self, path:str) -> None:
        for callback in self._events["OnScenarioSave"]._callbacks:
            callback(path)

    def _on_log_message(self, message:str, msgType:int, errorCode:int, fileName:str, lineNo:int, dispID:int) -> None:
        for callback in self._events["OnLogMessage"]._callbacks:
            callback(message, AgTypeNameMap["LogMessageType"](msgType), errorCode, fileName, lineNo, AgTypeNameMap["LogMessageDisplayID"](dispID))

    def _on_anim_update(self, timeEpSec:float) -> None:
        for callback in self._events["OnAnimUpdate"]._callbacks:
            callback(timeEpSec)

    def _on_new_globe_ctrl_request(self, SceneID:int) -> None:
        for callback in self._events["OnNewGlobeCtrlRequest"]._callbacks:
            callback(SceneID)

    def _on_new_map_ctrl_request(self, WinID:int) -> None:
        for callback in self._events["OnNewMapCtrlRequest"]._callbacks:
            callback(WinID)

    def _on_before_new_scenario(self, Scenario:str) -> None:
        for callback in self._events["OnBeforeNewScenario"]._callbacks:
            callback(Scenario)

    def _on_before_load_scenario(self, Scenario:str) -> None:
        for callback in self._events["OnBeforeLoadScenario"]._callbacks:
            callback(Scenario)

    def _on_begin_scenario_close(self) -> None:
        for callback in self._events["OnBeginScenarioClose"]._callbacks:
            callback()

    def _on_new_gfx_analysis_ctrl_request(self, SceneID:int, GfxAnalysisMode:int) -> None:
        for callback in self._events["OnNewGfxAnalysisCtrlRequest"]._callbacks:
            callback(SceneID, AgTypeNameMap["Graphics2DAnalysisMode"](GfxAnalysisMode))

    def _on_ssl_certificate_server_error(self, pArgs:"STKXSSLCertificateErrorEventArgs") -> None:
        for callback in self._events["OnSSLCertificateServerError"]._callbacks:
            callback(pArgs)

    def _on_con_control_quit_received(self, pArgs:"STKXConControlQuitReceivedEventArgs") -> None:
        for callback in self._events["OnConControlQuitReceived"]._callbacks:
            callback(pArgs)


################################################################################
#          IStkGraphicsSceneEvents
################################################################################

class ISceneEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgStkGraphicsSceneEvents, events)
        self._register_iadd_isub_callbacks("Rendering", self._rendering)

    def _rendering(self, Sender:typing.Any, Args:"RenderingEventArgs") -> None:
        for callback in self._events["Rendering"]._callbacks:
            callback(Sender, Args)


################################################################################
#          IStkGraphicsKmlGraphicsEvents
################################################################################

class IKmlGraphicsEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgStkGraphicsKmlGraphicsEvents, events)
        self._register_iadd_isub_callbacks("DocumentLoaded", self._document_loaded)

    def _document_loaded(self, Sender:typing.Any, Args:"KmlDocumentLoadedEventArgs") -> None:
        for callback in self._events["DocumentLoaded"]._callbacks:
            callback(Sender, Args)


################################################################################
#          IStkGraphicsImageCollectionEvents
################################################################################

class IImageCollectionEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgStkGraphicsImageCollectionEvents, events)
        self._register_iadd_isub_callbacks("AddComplete", self._add_complete)

    def _add_complete(self, Sender:typing.Any, Args:"GlobeImageOverlayAddCompleteEventArgs") -> None:
        for callback in self._events["AddComplete"]._callbacks:
            callback(Sender, Args)


################################################################################
#          IStkGraphicsTerrainCollectionEvents
################################################################################

class ITerrainOverlayCollectionEventGrpcHandler(GrpcEventHandlerImpl):

    def __init__(self, interface:GrpcInterface, events:dict):
        GrpcEventHandlerImpl.__init__(self, interface, AgGrpcServices_pb2.EventHandler.eIAgStkGraphicsTerrainCollectionEvents, events)
        self._register_iadd_isub_callbacks("AddComplete", self._add_complete)

    def _add_complete(self, Sender:typing.Any, Args:"TerrainOverlayAddCompleteEventArgs") -> None:
        for callback in self._events["AddComplete"]._callbacks:
            callback(Sender, Args)


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""The API migration assistant applies code transformations to upgrade code to a new version of an API."""

import datetime
import importlib.util
import logging
from pathlib import Path
import sys

from .argument_parser import ArgumentParser
from .code_editor import CodeEditor
from .mappings import Mappings
from .recorder import Recorder
from .recording import Recording


def record(
    mappings_directory,
    root_directory,
    recordings_directory,
    program,
    entry_point,
    run_as_module=False,
    program_args=None,
):
    """Invoke the specified entry point for the specified script or module, and record its execution."""
    mappings = Mappings.load_from(mappings_directory)
    if root_directory is None:
        if run_as_module:
            raise RuntimeError("root_directory argument is mandatory when running as a module")
        else:
            root_directory = str(Path(program).parent)
    recorder = Recorder(
        program,
        entry_point,
        root_directory,
        is_member_name_of_interest_func=mappings.is_member_name_of_interest,
        run_as_module=run_as_module,
        program_args=program_args,
    )
    recording = recorder.record()

    recordings_directory_path = Path(recordings_directory)
    date_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")[:-3]
    recordings_directory_path.mkdir(parents=True, exist_ok=True)
    recordings_file = recordings_directory_path / f"recording_{date_time}.json"
    recording.save(str(recordings_file.resolve()), f"{Path.cwd()}> {' '.join(sys.argv)}")


def _invoke_record(parser, args):
    """Invoke the record function above with the parameters extracted from the arguments."""
    if args.root_directory is None and args.run_as_module:
        parser.error("[--root-directory <directory>] must be specified when executing a module with [-m]")
    record(
        args.mappings_directory,
        args.root_directory,
        args.recordings_directory,
        args.program,
        args.entry_point,
        args.run_as_module,
        args.program_args,
    )


def apply(mappings_directory, recordings_directory):
    """Apply the migrations previously recorded."""
    logging.info(f"Applying changes from {recordings_directory}")
    if not importlib.util.find_spec("libcst"):
        logging.error("The libcst library is required. Please install it and re-run.")
    else:
        mappings = Mappings.load_from(mappings_directory)
        recording = Recording.load_from_recordings_in_directory(recordings_directory)
        editor = CodeEditor(recording, mappings)
        editor.apply_changes()


def _invoke_apply(parser, args):
    """
    Invoke the apply function above with the parameters extracted from the arguments.

    Callback passed to ``ArgumentParser.create_parser`` as its ``apply_handler`` parameter,
    which dictates the signature of the function, and which is why the ``parser`` argument
    is present even if not used in the function implementation.
    """
    apply(args.mappings_directory, args.recordings_directory)


def main(argv=None):
    """Parse the command line argument and invoke the corresponding action."""
    if argv is None:
        argv = sys.argv[1:]

    argument_parser = ArgumentParser.create_parser(record_handler=_invoke_record, apply_handler=_invoke_apply)
    args = argument_parser.parse_args(argv)
    _configure_logging(args)
    if hasattr(args, "handler"):
        args.handler(args)


def _configure_logging(args):
    """Configure logging as specified through the command line."""
    if args.log_level == "DEBUG":
        log_level = logging.DEBUG
    elif args.log_level == "WARNING":
        log_level = logging.WARNING
    elif args.log_level == "ERROR":
        log_level = logging.ERROR
    else:
        log_level = logging.INFO
    logging.basicConfig(level=log_level, format="%(levelname)s: %(message)s")


if __name__ == "__main__":
    sys.exit(main())
